import os
import pytest
import numpy as np
from leaf_diag.data_sync import SignalSynchronizer, SignalSynchronizerSingle, SignalSynchronizerMulti
from leaf_diag.data_loader import MotorCommandSignal, OrientationSignal, UlogMotorCommandSignal, load_rosbag, load_ulg
from leaf_diag.data_transform import px4_rpy_to_droneleaf_rpy
from leaf_diag.data_loader import (
    OrientationSignal,
    UlogOrientationSignal,
    CartesianSignal,
    UlogPositionSignal
)
import matplotlib.pyplot as plt


@pytest.fixture
def bag_data():
    data_dir = "data_test"
    filename = "2025-03-06-16-39-55.bag"
    path = os.path.join(data_dir, filename)
    topic_name = "/dynamo_db/actu_sys/arm_commander/io_cmds_to_px4"
    return {
        "path": path,
        "topic_name": topic_name,
        "motor_id": 3
    }

@pytest.fixture
def ulg_data():
    data_dir = "data_test"
    filename = "log_264_2025-3-5-12-17-16.ulg"
    path = os.path.join(data_dir, filename)
    ulog_name = "t_actuator_motors_0"
    motor_id = 3
    expected_offset = 5.796
    return {
        "path": path,
        "ulog_name": ulog_name,
        "motor_id": motor_id,
        "expected_offset": expected_offset
    }

def test_load_rosbag(bag_data):
    bag = load_rosbag(bag_data["path"])
    assert bag is not None

def test_load_ulg(ulg_data):
    ulg = load_ulg(ulg_data["path"])
    assert ulg is not None

def test_compute_best_offset(bag_data, ulg_data):
    sync = SignalSynchronizerSingle(bag_data["path"], ulg_data["path"])
    sync.compute_best_offset(
        ros_signal_class=MotorCommandSignal,
        ulog_signal_class=UlogMotorCommandSignal,
        ros_topic_name=bag_data["topic_name"],
        ulog_msg_key=ulg_data["ulog_name"],
        ros_signal_idx=bag_data["motor_id"],
        ulog_signal_idx=ulg_data["motor_id"]
    )
    assert np.isclose(sync.best_offset, ulg_data["expected_offset"], atol=0.1)

def test_compute_best_offset_all(bag_data, ulg_data):
    sync = SignalSynchronizerSingle(bag_data["path"], ulg_data["path"])
    sync.compute_best_offset(
        ros_signal_class=MotorCommandSignal,
        ulog_signal_class=UlogMotorCommandSignal,
        ros_topic_name=bag_data["topic_name"],
        ulog_msg_key=ulg_data["ulog_name"],
        ros_signal_idx=None,
        ulog_signal_idx=None
    )
    assert np.isclose(sync.best_offset, ulg_data["expected_offset"], atol=0.1)


def test_plot_correlation(bag_data, ulg_data):
    sync = SignalSynchronizerSingle(bag_data["path"], ulg_data["path"])
    sync.compute_best_offset(
        ros_signal_class=MotorCommandSignal,
        ulog_signal_class=UlogMotorCommandSignal,
        ros_topic_name=bag_data["topic_name"],
        ulog_msg_key=ulg_data["ulog_name"],
        ros_signal_idx=bag_data["motor_id"],
        ulog_signal_idx=ulg_data["motor_id"]
    )
    try:
        sync.plot_correlation(filename=os.path.join("data_test", "correlation.png"))
    except Exception as e:
        pytest.fail(f"plot_correlation raised an error: {e}")

def test_plot_signals(bag_data, ulg_data):
    sync = SignalSynchronizerSingle(bag_data["path"], ulg_data["path"])
    sync.compute_best_offset(
        ros_signal_class=MotorCommandSignal,
        ulog_signal_class=UlogMotorCommandSignal,
        ros_topic_name=bag_data["topic_name"],
        ulog_msg_key=ulg_data["ulog_name"],
        ros_signal_idx=bag_data["motor_id"],
        ulog_signal_idx=ulg_data["motor_id"]
    )
    try:
        sync.plot_signals(filename=os.path.join("data_test", "signals.png"))
    except Exception as e:
        pytest.fail(f"plot_signals raised an error: {e}")

def test_get_euler_angles(bag_data, ulg_data):
    sync = SignalSynchronizerSingle(bag_data["path"], ulg_data["path"])
    sync.compute_best_offset(
        ros_signal_class=MotorCommandSignal,
        ulog_signal_class=UlogMotorCommandSignal,
        ros_topic_name=bag_data["topic_name"],
        ulog_msg_key=ulg_data["ulog_name"],
        ros_signal_idx=bag_data["motor_id"],
        ulog_signal_idx=ulg_data["motor_id"]
    )

    with load_rosbag(bag_data["path"]) as bag:
        ros_ori_signal = OrientationSignal(bag, "/dynamo_db/estimator/ori")
        ros_times = ros_ori_signal.timestamps - np.min(ros_ori_signal.timestamps)
        ori = ros_ori_signal.data

    with load_ulg(ulg_data["path"]) as ulog:
        ulog_ori_signal = UlogOrientationSignal(ulog, "t_vehicle_attitude_0")
        ulog_times = ulog_ori_signal.timestamps - np.min(ulog_ori_signal.timestamps)
        RPY_px4 = ulog_ori_signal.data

    assert RPY_px4 is not None
    assert RPY_px4.shape[1] == 3

    # plot the Euler angles
    fig, ax = plt.subplots(3, 1, figsize=(10, 6), sharex=True)
    ax[0].plot(ulog_times + sync.best_offset, RPY_px4[:,0], label="Roll (PX4)")
    ax[0].plot(ros_times, ori[:,0], label="Roll (rosbag)")
    ax[0].set_ylabel("Roll (rad)")
    ax[0].legend()

    ax[1].plot(ulog_times + sync.best_offset, RPY_px4[:,1], label="Pitch (PX4)")
    ax[1].plot(ros_times, ori[:,1], label="Pitch (rosbag)")
    ax[1].set_ylabel("Pitch (rad)")
    ax[1].legend()

    ax[2].plot(ulog_times + sync.best_offset, RPY_px4[:,2], label="Yaw (PX4)")
    ax[2].plot(ros_times, ori[:,2], label="Yaw (rosbag)")
    ax[2].set_ylabel("Yaw (rad)")
    ax[2].legend()

    ax[2].set_xlabel("Time (s)")
    plt.tight_layout()
    fig.savefig(os.path.join("data_test", "angles.png"))


import time
import px4tools
from .ros_utils import MemoryBagReader
import pandas as pd

import numpy as np
from pyparsing import col

from .data_transform import quat_to_euler, px4_rpy_to_droneleaf_rpy
from contextlib import contextmanager

from typing import Tuple, Dict, Tuple, List, Generator
from .data_transform import compute_altitude_range, get_rotations_from_attitude
from scipy.spatial.transform import Rotation as R
import matplotlib.pyplot as plt
from scipy.signal import detrend
from .ros_utils import NamedBytesIO
import io

from typing import Union


@contextmanager
def load_rosbag(file_path: Union[str, bytes]) -> Generator[MemoryBagReader, None, None]:
    """
    Context manager for loading ROS bag files.
    Args:
        file_path (str): Path to the ROS bag file.
    Yields:
        MemoryBagReader: An instance of MemoryBagReader for the ROS bag file.
    Example:
        >>> with load_rosbag('path/to/your.bag') as bag:
        >>>     # Process the bag data
    Raises:
        FileNotFoundError: If the specified file does not exist.
        Exception: If there is an error while reading the bag file.
    """
    if isinstance(file_path, str):
        # If file_path is a string, we assume it's a file path
        bag = MemoryBagReader(file_path)
    elif isinstance(file_path, bytes):
        # If file_path is bytes, we assume it's a byte stream
        # Create a BytesIO object to read from the byte stream

        byte_stream = NamedBytesIO(file_path)
        byte_stream.seek(0)  # 2) ensure at start

        bag = MemoryBagReader(byte_stream)
    else:
        raise TypeError("file_path must be str or bytes")
    try:
        yield bag
    except Exception as e:
        raise Exception(f"Error reading ROS bag file: {e}")
    finally:
        # MemoryBagReader does not provide a close method, so no action is needed here
        pass

@contextmanager
def load_ulg(file_path: Union[str, bytes]) -> Generator[Dict[str, pd.DataFrame], None, None]:
    """
    Context manager for loading ULog files.
    Args:
        file_path (str): Path to the ULog file.
    Yields:
        dict: A dictionary containing the ULog data.
    Example:
        >>> with load_ulg('path/to/your.ulg') as ulg:
        >>>     # Process the ULog data
    Raises:
        FileNotFoundError: If the specified file does not exist.
        Exception: If there is an error while reading the ULog file.
    """
    if isinstance(file_path, str):
        ulg = px4tools.ulog.read_ulog(file_path)
    elif isinstance(file_path, bytes):
        # If file_path is bytes, we assume it's a byte stream
        # Create a BytesIO object to read from the byte stream
        byte_stream = io.BytesIO(file_path)
        ulg = px4tools.ulog.read_ulog(byte_stream)
    try:
        yield ulg
    finally:
        # px4tools.ulog does not require an explicit close, so no action is needed
        pass

def extract_ros_motor_signal(bag: MemoryBagReader, topic_name: str) -> Tuple[np.ndarray, np.ndarray]:
    """
    Extracts the ROS signal and its corresponding timestamps.
    Modify this function if your bag data structure is different.
    """
    # Example: assuming bag provides a dictionary with a 'time' key (absolute timestamps)
    # and the desired signal is at:
    # bag['arm_commander/io_cmds_to_px4/data/data'][0]
    # Extract the messages from the specific topic
    df_data = bag.message_by_topic_memory(topic_name)

    columns = []
    for motor_id in range(4):
        columns.append('data_%i' %(
            motor_id
        ))

    ros_signal = np.array(df_data[columns].values)
    ros_times = np.array(df_data['Time'].values)

    return ros_times, ros_signal

def extract_ulog_position_estimate_signal(ulog: dict, ulog_name: str, coordinate: str) -> Tuple[np.ndarray, np.ndarray]:
    """
    Extracts the Ulog signal and its timestamps.
    Modify this function if your Ulog structure is different.
    """
    # Example: assuming ulog provides a dictionary with a 'time' key (relative timestamps)
    # and the desired signal is in 'actuator_motors'
    ulog_times = np.array(ulog[ulog_name]["timestamp"])
    ulog_signal = np.array(ulog[ulog_name][coordinate])

    # convert ulog times to seconds
    ulog_times = ulog_times / 1e6

    return ulog_times, ulog_signal

def extract_ros_position_estimate_signal(bag: MemoryBagReader, topic_name: str, coordinate: str) -> Tuple[np.ndarray, np.ndarray]:
    """
    Extracts the ROS signal and its corresponding timestamps.
    Modify this function if your bag data structure is different.
    """
    # Example: assuming bag provides a dictionary with a 'time' key (absolute timestamps)
    # and the desired signal is at:
    # bag['arm_commander/io_cmds_to_px4/data/data'][0]
    # Extract the messages from the specific topic
    df_data = bag.message_by_topic_memory(topic_name)

    ros_signal = np.array(df_data[coordinate].values)
    ros_times = np.array(df_data['Time'].values)

    return ros_times, ros_signal

def extract_ulog_signal(ulog: dict, ulog_name: str) -> np.ndarray:
    """
    Extracts the Ulog signal and its timestamps.
    Modify this function if your Ulog structure is different.
    """
    # Example: assuming ulog provides a dictionary with a 'time' key (relative timestamps)
    # and the desired signal is in 'actuator_motors'

    return ulog[ulog_name]

def extract_ros_signal(bag: MemoryBagReader, topic_name: str) -> pd.DataFrame:
    """
    Extracts the ROS signal and its corresponding timestamps.
    Modify this function if your bag data structure is different.
    """
    # Example: assuming bag provides a dictionary with a 'time' key (absolute timestamps)
    # and the desired signal is at:
    # bag['arm_commander/io_cmds_to_px4/data/data'][0]
    # Extract the messages from the specific topic
    df_data = bag.message_by_topic_memory(topic_name)

    return df_data


class BaseSignal:
    def __init__(self, data, timestamps, dimension=None, data_keys=None):
        self.data = np.array(data)
        self.timestamps = np.array(timestamps)
        self.dimension = dimension
        self.data_keys = data_keys
        self.sampling_freq = self._compute_sampling_freq()

        self.reset_time()

    def _compute_sampling_freq(self):
        if len(self.timestamps) < 2:
            return 0
        diffs = np.diff(self.timestamps)
        return 1.0 / np.mean(diffs) if np.mean(diffs) != 0 else 0

    def set_start_time(self, start_time):
        self.start_time = start_time
        self.timestamps = self.timestamps - self.start_time

    def reset_time(self):
        self.start_time = self.timestamps[0]
        self.timestamps = self.timestamps - self.start_time

    def get_signal_slice(self, idx=None):
        return self.data if idx is None else self.data[:, idx]

    def plot(self, ax:plt.Axes=None, idx:int=None, title:str = None, label: str = None, **kwargs) -> Tuple[plt.Figure, plt.Axes]:
        """
        Plot the signal data.
        Args:
            ax: Matplotlib Axes object to plot on. If None, a new figure and axes are created.
            idx: Index of the signal to plot. If None, all signals are plotted.
            **kwargs: Additional keyword arguments for the plot.
        Returns:
            fig: Matplotlib Figure object.
            ax: Matplotlib Axes object.
        """
        if idx is None and self.dimension > 1:
            # Create subplots if we're plotting multiple dimensions
            fig, axes = plt.subplots(self.dimension, 1, figsize=(10, 2*self.dimension), sharex=True)
            colors = ['blue', 'red', 'green', 'purple', 'orange', 'brown', 'pink', 'gray', 'olive', 'cyan']
            for i in range(self.dimension):
                color_idx = i % len(colors)  # Cycle through colors if more dimensions than colors
                axes[i].plot(self.timestamps, self.data[:, i], color=colors[color_idx], **kwargs)
                if self.data_keys:
                    axes[i].set_ylabel(self.data_keys[i])
                if i == 0 and title:
                    axes[i].set_title(title, pad=10)  # Add vertical padding to title
            axes[-1].set_xlabel('Time (s)')
            fig.tight_layout()
            return fig, axes
        else:
            # Single plot for one dimension or specific index
            if ax is None:
                fig, ax = plt.subplots(figsize=(10, 5))
            else:
                fig = ax.figure
            
            if idx is None:
                for i in range(self.dimension):
                    if label is None:
                        label = self.data_keys[i] if self.data_keys else f"Signal {i}"
                    ax.plot(self.timestamps, self.data[:, i], label=label, **kwargs)
                ax.legend()
            else:
                if label is None:
                    label = self.data_keys[idx] if self.data_keys else f"Signal {idx}"
                ax.plot(self.timestamps, self.data[:, idx], label=label, **kwargs)
                ax.legend()
            
            if title:
                ax.set_title(title)
                ax.set_xlabel('Time (s)')
            return fig, ax

class RosSignal(BaseSignal):
    def __init__(self, ros_bag, topic, dimension=None, data_keys=None):
        # Example extraction from a ROS bag
        
        topic_data = extract_ros_signal(ros_bag, topic).to_numpy()
        timestamps = topic_data[:, 0]
        data = topic_data[:, 1:]

        self.start_time = timestamps[0]
        self.data = data
        self.dimension = data.shape[1]
        self.topic = topic

        super().__init__(data, timestamps, dimension, data_keys)


class UlogSignal(BaseSignal):
    def __init__(self, ulog: Dict[str,pd.DataFrame], msg_key: str, data_keys: List[str]=None, dimension: int=None):
        # Example extraction from a ULog
        df = ulog[msg_key]
        timestamps = df["timestamp"].to_numpy() / 1e6
        self.start_time = timestamps[0]
        self.msg_key = msg_key

        if data_keys is None:
            data = df.drop(columns=["timestamp"]).to_numpy()
        else:
            data = df[data_keys].to_numpy()

        if data.ndim == 1:
            data = data[:, None]

        if dimension is None:
            dimension = data.shape[1]
        super().__init__(data, timestamps, dimension, data_keys)


class MotorCommandSignal(RosSignal):
    def __init__(self, ros_bag: MemoryBagReader, topic: str):
        """
        Extracts motor command signals from a ROS bag.
        Args:
            ros_bag: The ROS bag reader object.
            topic: The topic name to extract motor commands from.
        """
        # Extract one motor channel
        times, values = extract_ros_motor_signal(ros_bag, topic)

        self.dimension = values.shape[1]
        self.data = values
        self.timestamps = times
        self.sampling_freq = self._compute_sampling_freq()

        self.data_keys = [f"motor_{i}" for i in range(self.dimension)]

        self.reset_time()


class UlogMotorCommandSignal(UlogSignal):
    def __init__(self, ulog, msg_key):
        super().__init__(ulog, msg_key, data_keys=[f"f_control_{i}_" for i in range(4)], dimension=4)


class OrientationSignal(RosSignal):
    def __init__(self, ros_bag, topic):
        data_keys = ["roll", "pitch", "yaw"]
        super().__init__(ros_bag, topic, dimension=3, data_keys=data_keys)


class UlogOrientationSignal(UlogSignal):
    def __init__(self, ulog, msg_key):
        df = ulog[msg_key]
        # Assume quaternion to Euler has been done externally
        times = df["timestamp"].to_numpy() / 1e6
        euler = quat_to_euler(df[["f_q_1_", "f_q_2_", "f_q_3_", "f_q_0_"]].to_numpy())
        euler = px4_rpy_to_droneleaf_rpy(euler)

        self.dimension = euler.shape[1]
        self.data = euler
        self.timestamps = times
        self.sampling_freq = self._compute_sampling_freq()

        self.data_keys = ["roll", "pitch", "yaw"]

        self.reset_time()


class UlogAngularSignal(UlogSignal):

    def __init__(self, ulog, msg_key):
        super().__init__(ulog, msg_key, data_keys=["f_xyz_0_", "f_xyz_1_", "f_xyz_2_"], dimension=3)
        self.transform_to_droneleaf()

    def transform_to_droneleaf(self):
        x_px4 = self.data[:, 0].copy()
        y_px4 = self.data[:, 1].copy()
        z_px4 = self.data[:, 2].copy()

        self.data[:, 0] = x_px4
        self.data[:, 1] = -y_px4
        self.data[:, 2] = -z_px4


class CartesianSignal(RosSignal):
    def __init__(self, ros_bag, topic):
        data_keys = ["x", "y", "z"]
        super().__init__(ros_bag, topic, dimension=3, data_keys=data_keys)


class UlogCartesianSignal(UlogSignal):

    def get_initial_coords(self):
        return self.data[0]

    def transform_to_droneleaf(self):
        x_px4 = self.data[:, 0].copy()
        y_px4 = self.data[:, 1].copy()
        z_px4 = self.data[:, 2].copy()

        self.data[:, 0] = y_px4
        self.data[:, 1] = x_px4
        self.data[:, 2] = -z_px4


class UlogPositionSignal(UlogCartesianSignal):
    def __init__(self, ulog, msg_key):
        super().__init__(ulog, msg_key, data_keys=["f_x", "f_y", "f_z"], dimension=3)
        self.transform_to_droneleaf()


class UlogVelocitySignal(UlogCartesianSignal):
    def __init__(self, ulog, msg_key):
        super().__init__(ulog, msg_key, data_keys=["f_vx", "f_vy", "f_vz"], dimension=3)
        self.transform_to_droneleaf()


class UlogAccelerationSignal(UlogCartesianSignal):
    def __init__(self, ulog, msg_key):
        super().__init__(ulog, msg_key, data_keys=["f_ax", "f_ay", "f_az"], dimension=3)
        self.transform_to_droneleaf()

class UlogRawAccelerationSignal(UlogCartesianSignal):
    def __init__(self, ulog, msg_key):
        super().__init__(ulog, msg_key, data_keys=["f_x", "f_y", "f_z"], dimension=3)
        self.transform_to_droneleaf()

    def transform_to_droneleaf(self):
        x_px4 = self.data[:, 0].copy()
        y_px4 = self.data[:, 1].copy()
        z_px4 = self.data[:, 2].copy()

        self.data[:, 0] = x_px4
        self.data[:, 1] = -y_px4
        self.data[:, 2] = -z_px4 - 9.81

class ULogGPSPositionSignal(UlogCartesianSignal):
    def __init__(self, ulog, msg_key):
        super().__init__(ulog, msg_key, data_keys=["f_lat", "f_lon", "f_alt"], dimension=3)
        self.initial_coords: Tuple[float, float, float] = None
        # Convert GPS coordinates to local x, y, z
        self._convert_gps_to_local()

    def _convert_gps_to_local(self):
        """Convert GPS coordinates to local x, y, z coordinates"""
        # Unpack initial coordinates
        
        # Extract GPS data
        lats = self.data[:, 0].copy()
        lons = self.data[:, 1].copy()
        alts = self.data[:, 2].copy()
        
        self.initial_coords = (lats[0], lons[0], alts[0])

        # Constants for conversion
        R_EARTH = 6378137.0  # Earth radius in meters
        
        # Convert lat/lon to meters (using equirectangular approximation)
        # This is a simple approximation valid for small distances
        x = R_EARTH * np.cos(np.radians(self.initial_coords[0])) * np.radians(lons - self.initial_coords[1])
        y = R_EARTH * np.radians(lats - self.initial_coords[0])
        z = alts - self.initial_coords[2]
        
        # Replace GPS data with local coordinates
        self.data[:, 0] = x
        self.data[:, 1] = y
        self.data[:, 2] = z
    
    def apply_calibration_offset(self, calibration_offset: Tuple[float, float, float]):
        """
        Apply calibration offsets to the x, y, z position data.
        
        Args:
            calibration_offset: A tuple of (x_offset, y_offset, z_offset) in meters
        """
        if len(calibration_offset) != 3:
            raise ValueError("Calibration offset must be a tuple of 3 values (x, y, z)")
            
        # Apply offsets to the data
        self.data[:, 0] += calibration_offset[0]  # x offset
        self.data[:, 1] += calibration_offset[1]  # y offset 
        self.data[:, 2] += calibration_offset[2]  # z offset

class UlogGPSVelocitySignal(UlogCartesianSignal):
    def __init__(self, ulog, msg_key):
        super().__init__(ulog, msg_key, data_keys=["f_vel_e_m_s", "f_vel_n_m_s", "f_vel_d_m_s"], dimension=3)
        self.transform_to_droneleaf()

    def transform_to_droneleaf(self):
        x_px4 = self.data[:, 0].copy()
        y_px4 = self.data[:, 1].copy()
        z_px4 = self.data[:, 2].copy()

        self.data[:, 0] = x_px4
        self.data[:, 1] = y_px4
        self.data[:, 2] = -z_px4

class UlogOFVelocity(UlogCartesianSignal):
    def __init__(self, ulog, of_msg_key, pos_msg_key, att_msg_key, distance_sensor_msg_key):
        """
        Calculate ground velocity from optical flow data.
        
        Args:
            ulog: ULog data dictionary
            of_msg_key: Key for optical flow message
            pos_msg_key: Key for vehicle position message (for altitude)
        """
        # Create derived optical flow velocity dataset
        of_velocity_df = self._calculate_flow_velocity(ulog, of_msg_key, pos_msg_key, att_msg_key, distance_sensor_msg_key)
        
        # Add to ulog dictionary under a new key
        msg_key = "optical_flow_velocity"
        ulog[msg_key] = of_velocity_df
        
        # Initialize using the parent class with our derived data
        super().__init__(ulog, msg_key, data_keys=["f_vx", "f_vy", "f_vz"], dimension=2)
        
        # Transform to DroneLeaf coordinate system
        self.transform_to_droneleaf()
    
    def transform_to_droneleaf(self):
        x_px4 = self.data[:, 0].copy()
        y_px4 = self.data[:, 1].copy()
        z_px4 = self.data[:, 2].copy()

        self.data[:, 0] = -y_px4
        self.data[:, 1] = x_px4
        self.data[:, 2] = -z_px4

    def _calculate_flow_velocity(self, ulog, of_msg_key, pos_msg_key, att_msg_key, distance_sensor_msg_key):
        """
        Calculate velocity from optical flow, attitude, and altitude data,
        replicating the PX4 EKF sign flips and rotation compensation.
        Uses compute_altitude_range for accurate range measurement.
        """
        # -------------------------------------------------------------------
        # 1) Extract optical flow data
        #    Expect columns: 'timestamp', 'pixel_flow[0]', 'pixel_flow[1]',
        #                   'delta_angle[0]', 'delta_angle[1]', 'delta_angle[2]',
        #                   'integration_timespan_us', 'quality', ...
        #    The exact column names can differ in your logs:
        # -------------------------------------------------------------------
        of_df = ulog[of_msg_key].copy()

        # For example, rename to simpler columns:
        of_df.rename(columns={
            "f_pixel_flow_0_": "px_flow_x",
            "f_pixel_flow_1_": "px_flow_y",
            "f_delta_angle_0_": "dangle_x",
            "f_delta_angle_1_": "dangle_y",
            "f_delta_angle_2_": "dangle_z",
            "f_integration_timespan_us": "dt_us"
        }, inplace=True)

        # Make sure we have the needed columns
        needed_cols = ["timestamp", "px_flow_x", "px_flow_y", 
                       "dangle_x", "dangle_y", "dangle_z", "dt_us"]
        for c in needed_cols:
            if c not in of_df.columns:
                raise ValueError(f"Missing column '{c}' in optical flow dataset")

        # Convert timestamp to float seconds
        of_df["t"] = of_df["timestamp"] * 1e-6
        # Convert dt from microseconds to seconds
        of_df["dt"] = of_df["dt_us"] * 1e-6

        # -------------------------------------------------------------------
        # 2) Extract position and attitude data for altitude calculation
        # -------------------------------------------------------------------
        pos_df = ulog[pos_msg_key].copy()
        # Expect columns: 'timestamp', 'f_x', 'f_y', 'f_z', etc.
        pos_df.rename(columns={"f_x": "x_ned", "f_y": "y_ned", "f_z": "z_ned"}, inplace=True)
        pos_df["t"] = pos_df["timestamp"] * 1e-6

        att_df = ulog[att_msg_key].copy()
        # Typically has columns: 'timestamp', 'q[0]', 'q[1]', 'q[2]', 'q[3]'
        att_df.rename(columns={
            "f_q_0_": "qw",
            "f_q_1_": "qx",
            "f_q_2_": "qy",
            "f_q_3_": "qz"
        }, inplace=True)
        att_df["t"] = att_df["timestamp"] * 1e-6

        # We'll do an asof merge to get attitude per flow sample
        # Sort first:
        of_df.sort_values("t", inplace=True)
        att_df.sort_values("t", inplace=True)
        pos_df.sort_values("t", inplace=True)

        # Merge attitude data with optical flow data
        merged = pd.merge_asof(of_df, att_df[["t","qw","qx","qy","qz"]], 
                               on="t", direction="nearest")
        
        # -------------------------------------------------------------------
        # 3) Calculate accurate altitude range using compute_altitude_range
        # -------------------------------------------------------------------
        
        n = len(merged)
        # Create rotation matrices for each sample
        R_to_earth = np.zeros((3, 3, n))
        state_pos = np.zeros((3, n))
        
        # Interpolate position at the OF timestamps
        of_times = merged["t"].to_numpy()
        pos_times = pos_df["t"].to_numpy()
        
        # Extract NED positions
        state_pos[0, :] = np.interp(of_times, pos_times, pos_df["x_ned"].to_numpy())
        state_pos[1, :] = np.interp(of_times, pos_times, pos_df["y_ned"].to_numpy())
        state_pos[2, :] = np.interp(of_times, pos_times, pos_df["z_ned"].to_numpy())
        
        # Calculate rotation matrices for each timestamp
        for i in range(n):
            _, rot_body_to_ned = get_rotations_from_attitude(
                merged["qw"].iloc[i],
                merged["qx"].iloc[i],
                merged["qy"].iloc[i],
                merged["qz"].iloc[i]
            )
            R_to_earth[:, :, i] = rot_body_to_ned.as_matrix()
        
        # # Compute accurate altitude range
        # range_to_ground = compute_altitude_range(
        #     R_to_earth=R_to_earth,
        #     state_pos=state_pos
        # )

        distance_df = ulog[distance_sensor_msg_key].copy()
        distance_df["t"] = distance_df["timestamp"] * 1e-6
        distance_df.rename(columns={"f_current_distance": "sensor_range"}, inplace=True)
        distance_df.sort_values("t", inplace=True)

        merged = pd.merge_asof(
            merged.sort_values("t"),
            distance_df[["t", "sensor_range"]].sort_values("t"),
            on="t",
            direction="nearest"
        )

        merged.rename(columns={"sensor_range": "range"}, inplace=True)

        # -------------------------------------------------------------------
        # 4) Replicate the sign flipping & subtract gyro
        # -------------------------------------------------------------------
        # flow_xy_rad  = -pixel_flow[0..1]
        merged["flow_x_rad"] = -merged["px_flow_x"]
        merged["flow_y_rad"] = -merged["px_flow_y"]

        # gyro_xyz = -delta_angle[0..2]
        merged["gyro_x"] = -merged["dangle_x"]
        merged["gyro_y"] = -merged["dangle_y"]
        # merged["gyro_z"] = -merged["dangle_z"]  # not used for XY flow

        # compensated_flow_xy = flow_xy_rad - gyro_xyz.xy()
        merged["compensated_flow_x"] = merged["flow_x_rad"] - merged["gyro_x"]
        merged["compensated_flow_y"] = merged["flow_y_rad"] - merged["gyro_y"]

        # rad/s
        merged["of_rate_x"] = merged["compensated_flow_x"] / merged["dt"]
        merged["of_rate_y"] = merged["compensated_flow_y"] / merged["dt"]

        # -------------------------------------------------------------------
        # 5) Convert to velocity in BODY frame using accurate range
        #    v_bx = -of_rate_y * range
        #    v_by =  of_rate_x * range
        #    v_bz =  0
        # -------------------------------------------------------------------
        merged["v_bx"] = -merged["of_rate_y"] * merged["range"]
        merged["v_by"] = -merged["of_rate_x"] * merged["range"]
        merged["v_bz"] =  0.0

        # -------------------------------------------------------------------
        # 6) Rotate body->NED for final velocity
        # -------------------------------------------------------------------
        vx_ned = []
        vy_ned = []
        vz_ned = []

        
        for i, row in merged.iterrows():
            # Build the Rotation from NED->body
            rot_ned_to_body = R.from_quat([row["qx"], row["qy"], row["qz"], row["qw"]])
            rot_body_to_ned = rot_ned_to_body.inv()

            v_body = np.array([row["v_bx"], row["v_by"], row["v_bz"]])
            v_ned  = rot_body_to_ned.apply(v_body)

            vx_ned.append(v_ned[0])
            vy_ned.append(v_ned[1])
            vz_ned.append(v_ned[2])

        merged["f_vx"] = vx_ned
        merged["f_vy"] = vy_ned
        merged["f_vz"] = vz_ned

        # -------------------------------------------------------------------
        # 7) Create and return the final DF
        # -------------------------------------------------------------------
        result_df = pd.DataFrame()
        # Use original 'timestamp' in microseconds
        result_df["timestamp"] = merged["timestamp"].astype(int)
        result_df["f_vx"] = merged["f_vx"]
        result_df["f_vy"] = merged["f_vy"]
        result_df["f_vz"] = merged["f_vz"]

        return result_df
    
class UlogHeading(UlogSignal):
    def __init__(self, ulog, mag_sensor_msg_key, att_msg_key):
        """
        Calculate ground velocity from optical flow data.
        
        Args:
            ulog: ULog data dictionary
            of_msg_key: Key for optical flow message
            pos_msg_key: Key for vehicle position message (for altitude)
        """
        # Create derived optical flow velocity dataset
        heading_df = self._calculate_heading(ulog, mag_sensor_msg_key, att_msg_key)
        
        # Add to ulog dictionary under a new key
        msg_key = "calculated_heading"
        ulog[msg_key] = heading_df
        
        # Initialize using the parent class with our derived data
        super().__init__(ulog, msg_key, data_keys=["f_heading",])

    def _calculate_heading(self, ulog, mag_sensor_msg_key, att_msg_key):
        """
        Calculate velocity from magnetometer flow, attitude, and altitude data,
        replicating the PX4 EKF sign flips and rotation compensation.
        Uses compute_altitude_range for accurate range measurement.
        """

        # -------------------------------------------------------------------
        # 1) Extract magnetometer and attitude data for heading calculation
        # -------------------------------------------------------------------
        mag_df = ulog[mag_sensor_msg_key].copy()
        # Expect columns: 'timestamp', 'f_x', 'f_y', 'f_z', etc.
        mag_df.rename(columns={"f_x": "f_x", "f_y": "f_y", "f_z": "f_z"}, inplace=True)
        mag_df["t"] = mag_df["timestamp"] * 1e-6

        att_df = ulog[att_msg_key].copy()
        # Typically has columns: 'timestamp', 'q[0]', 'q[1]', 'q[2]', 'q[3]'
        att_df.rename(columns={
            "f_q_0_": "qw",
            "f_q_1_": "qx",
            "f_q_2_": "qy",
            "f_q_3_": "qz"
        }, inplace=True)
        att_df["t"] = att_df["timestamp"] * 1e-6

        # We'll do an asof merge to get attitude per flow sample
        # Sort first:
        att_df.sort_values("t", inplace=True)
        mag_df.sort_values("t", inplace=True)

        # Merge attitude data with optical flow data
        merged = pd.merge_asof(mag_df, att_df[["t","qw","qx","qy","qz"]], 
                               on="t", direction="nearest")
        
        # -------------------------------------------------------------------
        # 3) Calculate roll pitch and yaw from quaternion in dataframe
        # -------------------------------------------------------------------
        
        # Calculate roll, pitch, yaw from quaternion
        euler = quat_to_euler(
            merged[["qw", "qx", "qy", "qz"]].to_numpy()
        )

        merged["roll"] = euler[:, 0] # roll in radians
        merged["pitch"] = euler[:, 1] # pitch in radians
        merged["yaw"] = euler[:, 2] # yaw in radians
        
        # -------------------------------------------------------------------
        # 4) Calculate magnetometer heading (yaw) with inclination compensation
        # -------------------------------------------------------------------

        # Extract magnetometer data
        mag_x = merged["f_x"].to_numpy()
        mag_y = merged["f_y"].to_numpy()
        mag_z = merged["f_z"].to_numpy()

        # Interpolate roll and pitch to magnetometer timestamps
        mag_timestamps = merged["t"].to_numpy()
        att_timestamps = merged["t"].to_numpy()

        roll_raw = merged["roll"].to_numpy()
        pitch_raw = merged["pitch"].to_numpy()

        # Use numpy.interp for interpolation
        roll_interp = np.interp(mag_timestamps, att_timestamps, roll_raw)
        pitch_interp = np.interp(mag_timestamps, att_timestamps, pitch_raw)

        roll = np.radians(roll_interp)  # roll in radians
        pitch = np.radians(pitch_interp)  # pitch in radians

        # Compensate for inclination
        Xh = mag_x * np.cos(pitch) + mag_z * np.sin(pitch)
        Yh = mag_x * np.sin(roll) * np.sin(pitch)
        Yh += mag_y * np.cos(roll) - mag_z * np.sin(roll) * np.cos(pitch)

        heading_rad = np.arctan2(Yh, Xh)
        heading_deg = np.degrees(heading_rad)
        merged["f_heading"] = heading_deg

        # -------------------------------------------------------------------
        # 5) Create and return the final DF
        # -------------------------------------------------------------------
        result_df = pd.DataFrame()
        # Use original 'timestamp' in microseconds
        result_df["timestamp"] = merged["timestamp"].astype(int)
        result_df["f_heading"] = merged["f_heading"]

        return result_df

class UlogEstimatorOFVelocity(UlogSignal):
    def __init__(self, ulog, msg_key):
        super().__init__(ulog, msg_key, data_keys=["f_vel_ne_0_", "f_vel_ne_1_"], dimension=2)
        self.transform_to_droneleaf()
    
    def transform_to_droneleaf(self):
        x_px4 = self.data[:, 0].copy()
        y_px4 = self.data[:, 1].copy()

        self.data[:, 0] = y_px4
        self.data[:, 1] = x_px4

class UlogGyroHeading(UlogSignal):
    def __init__(self, ulog, gyro_msg_key, attitude_msg_key):
        """
        Initialize the UlogGyroHeading class.

        Args:
            ulog: ULog data dictionary.
            gyro_msg_key: Key for the sensor_gyro message.
            attitude_msg_key: Key for the vehicle_attitude message.
        """
        # Calculate integrated angular displacement
        integrated_angles_df = self._calculate_integrated_angles(ulog, gyro_msg_key, attitude_msg_key)

        # Add to ulog dictionary under a new key
        msg_key = "sensor_gyro_integrated"
        ulog[msg_key] = integrated_angles_df

        # Initialize using the parent class with our derived data
        super().__init__(ulog, msg_key, data_keys=["angle_x", "angle_y", "angle_z"], dimension=3)

    def _calculate_integrated_angles(self, ulog, gyro_msg_key, attitude_msg_key):
        """
        Calculate integrated angular displacement from angular velocity data.

        Args:
            ulog: ULog data dictionary.
            gyro_msg_key: Key for the sensor_gyro message.
            attitude_msg_key: Key for the vehicle_attitude message.

        Returns:
            DataFrame with columns for timestamp, angle_x, angle_y, and angle_z.
        """
        # Extract angular velocity data
        gyro_df = ulog[gyro_msg_key]
        timestamps = gyro_df["timestamp"].to_numpy() / 1e6  # Convert to seconds
        angular_velocity = gyro_df[["f_x", "f_y", "f_z"]].to_numpy()

        # --- Bias Correction ---
        # Estimate bias from stationary data (e.g., first 100 samples)
        bias = np.mean(angular_velocity[:100], axis=0)
        angular_velocity -= bias

        # --- Detrending ---
        angular_velocity = detrend(angular_velocity, axis=0)

        # Compute time deltas
        time_deltas = np.diff(timestamps, prepend=timestamps[0])

        # Integrate angular velocity to get angular displacement
        integrated_angles = np.cumsum(angular_velocity * time_deltas[:, None], axis=0)

        # Extract initial attitude from vehicle_attitude topic
        attitude_df = ulog[attitude_msg_key]
        initial_quaternion = attitude_df.iloc[0][["f_q_1_", "f_q_2_", "f_q_3_", "f_q_0_"]].to_numpy()
        initial_euler = quat_to_euler(initial_quaternion[None, :])[0]  # Convert to roll, pitch, yaw

        # Add initial angles as constants of integration
        integrated_angles[:, 0] += initial_euler[0]  # Roll
        integrated_angles[:, 1] += initial_euler[1]  # Pitch
        integrated_angles[:, 2] += initial_euler[2]  # Yaw

        # Create a DataFrame for the integrated angles
        integrated_angles_df = pd.DataFrame({
            "timestamp": timestamps * 1e6,  
            "angle_x": integrated_angles[:, 0],
            "angle_y": integrated_angles[:, 1],
            "angle_z": integrated_angles[:, 2]
        })

        return integrated_angles_df


import os
from matplotlib import lines
import numpy as np

import matplotlib.pyplot as plt
from pyparsing import line

from .data_loader import (
    load_rosbag,
    load_ulg,
    RosSignal,
    MotorCommandSignal,
    UlogMotorCommandSignal,
    CartesianSignal,
    UlogSignal,
    OrientationSignal,
    UlogOrientationSignal,
    UlogAccelerationSignal,
    UlogCartesianSignal

)

from typing import Union, List
from scipy.interpolate import interp1d
from tqdm import tqdm

def _get_signal_nan_indices(signal):
    if signal.ndim > 1:
        return ~np.isnan(signal).any(axis=1)
    else:
        return ~np.isnan(signal)

def resample_signal(orig_times, signal, new_times):
    """
    Resamples the given signal onto the new time base using linear interpolation.
    Values outside the original time range will be nan.

    signal can be a 1D array or a 2D array (time x channels).
    """
    
    # loop over channels if signal is 2D
    if signal.ndim > 1:
        resampled_signal = np.zeros((len(new_times), signal.shape[1]))
        for i in range(signal.shape[1]):
            interp_func = interp1d(orig_times, signal[:,i], kind='linear', bounds_error=False, fill_value=np.nan)
            resampled_signal[:,i] = interp_func(new_times)
        return resampled_signal
    else:
        interp_func = interp1d(orig_times, signal, kind='linear', bounds_error=False, fill_value=np.nan)
        return interp_func(new_times)

def get_common_time_base(ros_times, ulog_times):
    """
    Compute the common time base for ROS and Ulog signals.
    """
    fs_ros = 1 / np.mean(np.diff(ros_times))
    fs_ulog = 1 / np.mean(np.diff(ulog_times))
    fs = max(fs_ros, fs_ulog)
    t_min, t_max = min(ros_times[0], ulog_times[0]), max(ros_times[-1], ulog_times[-1])
    common_times = np.arange(t_min, t_max, 1 / fs)

    return common_times, fs


class SignalSynchronizer:
    def __init__(self, bag_path: Union[str, bytes], ulg_path: Union[str, bytes]):
        self.bag_path = bag_path
        self.ulg_path = ulg_path
        self.common_times = None
        self.best_offset = None
        self.best_overlap = None
        self.overlaps = None
        self.correlations = None
        self.fs = None

    def _compute_best_offset(self, ros_times, ulog_times, ros_signal, ulog_signal):

        # Common time base
        self.common_times, self.fs = get_common_time_base(ros_times, ulog_times)

        # Resample
        self.ros_signal_resampled = resample_signal(ros_times, ros_signal, self.common_times)
        self.ulog_signal_resampled = resample_signal(ulog_times, ulog_signal, self.common_times)

        # Correlation
        signals_dict = {
            'ROS': {
                "length": len(self.ros_signal_resampled[_get_signal_nan_indices(self.ros_signal_resampled)]),
                "signal": self.ros_signal_resampled
            },
            'Ulog': {
                "length": len(self.ulog_signal_resampled[_get_signal_nan_indices(self.ulog_signal_resampled)]),
                "signal": self.ulog_signal_resampled
            }
        }
        self.shorter_type = min(signals_dict, key=lambda x: signals_dict[x]['length'])
        self.longer_type = max(signals_dict, key=lambda x: signals_dict[x]['length'])

        self.correlations = []
        
        # Extend search range to include negative offsets
        max_positive_offset = signals_dict[self.longer_type]["length"] + signals_dict[self.shorter_type]["length"]
        max_negative_offset = signals_dict[self.shorter_type]["length"]
        
        # Search from -max_negative_offset to max_positive_offset
        offset_range = range(-max_negative_offset, max_positive_offset + 1)

        corr_signal_short_clean = signals_dict[self.shorter_type]["signal"][
            _get_signal_nan_indices(signals_dict[self.shorter_type]["signal"])
        ]
        corr_signal_long_clean = signals_dict[self.longer_type]["signal"][
            _get_signal_nan_indices(signals_dict[self.longer_type]["signal"])
        ]

        self.overlaps = []

        for offset in tqdm(offset_range, desc="Calculating correlation"):
            if offset >= 0:
                # Positive offset: shorter signal starts after longer signal
                start = min(offset, len(corr_signal_long_clean))
                end = min(offset + len(corr_signal_short_clean), len(corr_signal_long_clean))

                corr_signal_long = corr_signal_long_clean[start : end]
                corr_signal_short = corr_signal_short_clean[:len(corr_signal_long)]

                self.overlaps.append((end - start) / len(corr_signal_short_clean))
            else:
                # Negative offset: shorter signal starts before longer signal
                abs_offset = abs(offset)
                start = abs_offset
                end = len(corr_signal_short_clean)

                corr_signal_short = corr_signal_short_clean[start : end]
                corr_signal_long = corr_signal_long_clean[:len(corr_signal_short)]

                self.overlaps.append((end - start) / len(corr_signal_short_clean))

            # Skip if either signal is empty
            if len(corr_signal_short) == 0 or len(corr_signal_long) == 0:
                self.correlations.append(np.nan)
                continue

            # Check if signals are multi-dimensional
            if corr_signal_short.ndim > 1 and corr_signal_long.ndim > 1:
                # Compute correlations dimension by dimension and average
                corrs = []
                for dim in range(corr_signal_short.shape[1]):
                    # Get data for this dimension
                    short_dim = corr_signal_short[:, dim]
                    long_dim = corr_signal_long[:, dim]
                    
                    # Find valid (non-NaN) indices for both signals
                    valid_indices = ~(np.isnan(short_dim) | np.isnan(long_dim))
                    if np.sum(valid_indices) < 2:  # Need at least 2 points for correlation
                        corrs.append(np.nan)
                    else:
                        short_valid = short_dim[valid_indices]
                        long_valid = long_dim[valid_indices]
                        
                        # Check if signals have enough variance
                        if np.std(short_valid) == 0 or np.std(long_valid) == 0:
                            corrs.append(np.nan)
                        else:
                            dim_corr = np.corrcoef(short_valid, long_valid)
                            corrs.append(dim_corr[0, 1])
                
                # Only compute mean over valid correlations
                valid_corrs = [c for c in corrs if not np.isnan(c)]
                if valid_corrs:
                    self.correlations.append(np.mean(valid_corrs))
                else:
                    self.correlations.append(np.nan)
            else:
                # For 1D signals, handle NaNs
                valid_indices = ~(np.isnan(corr_signal_short) | np.isnan(corr_signal_long))

                if np.sum(valid_indices) < 2:  # Need at least 2 points for correlation
                    self.correlations.append(np.nan)
                else:
                    short_valid = corr_signal_short[valid_indices]
                    long_valid = corr_signal_long[valid_indices]
                    
                    # Check if signals have enough variance
                    if np.std(short_valid) == 0 or np.std(long_valid) == 0:
                        self.correlations.append(np.nan)
                    else:
                        corr = np.corrcoef(short_valid, long_valid)
                        self.correlations.append(corr[0, 1])

        # Filter correlations by overlap threshold
        valid_correlations = np.array(self.correlations)
        overlap_mask = np.array(self.overlaps) > 0.5
        valid_correlations[~overlap_mask] = np.nan
        
        best_idx = np.nanargmax(valid_correlations)

        # Convert index back to actual offset (accounting for negative range)
        self.best_offset = (offset_range[best_idx]) / self.fs
        self.best_overlap = self.overlaps[best_idx]

    def plot_correlation(self, filename=None):
        if self.correlations is None:
            print("No correlation data. Run compute_best_offset first.")
            return
        
        # Recreate the offset range to match what was used in compute_best_offset
        signals_dict = {
            'ROS': {
                "length": len(self.ros_signal_resampled[_get_signal_nan_indices(self.ros_signal_resampled)]),
            },
            'Ulog': {
                "length": len(self.ulog_signal_resampled[_get_signal_nan_indices(self.ulog_signal_resampled)]),
            }
        }
        shorter_type = min(signals_dict, key=lambda x: signals_dict[x]['length'])
        longer_type = max(signals_dict, key=lambda x: signals_dict[x]['length'])
        
        max_positive_offset = signals_dict[longer_type]["length"] + signals_dict[shorter_type]["length"]
        max_negative_offset = signals_dict[shorter_type]["length"]
        offset_range = range(-max_negative_offset, max_positive_offset + 1)
        
        lags = np.array(offset_range) / self.fs
        plt.figure()
        plt.plot(lags, self.correlations)
        plt.axvline(self.best_offset, color='r', linestyle='--',
                    label=f'Best offset: {self.best_offset:.3f}s')
        plt.xlabel("Lag (seconds)")
        plt.ylabel("Cross correlation")
        plt.title("Cross correlation")
        plt.legend()
        if filename:
            plt.savefig(filename)
        else:
            plt.show()
        plt.close()

        overlaps = np.array(self.overlaps)
        plt.figure()
        plt.plot(lags, overlaps)
        plt.axvline(self.best_offset, color='r', linestyle='--',
                    label=f'Best offset: {self.best_offset:.3f}s')
        plt.xlabel("Lag (seconds)")
        plt.ylabel("Overlap")
        plt.title("Overlap")
        plt.legend()
        if filename:
            plt.savefig(filename.replace('.png', '_overlap.png'))
        else:
            plt.show()
        plt.close()

    def plot_angles(self, RPY_leaf, RPY_px4, filename=None):
        if RPY_leaf is None or RPY_px4 is None:
            print("No angle data. Run get_euler_angles first.")
            return
        t = np.arange(RPY_leaf.shape[1]) / self.fs
        plt.figure()
        for i, label in enumerate(['Roll', 'Pitch', 'Yaw']):
            plt.subplot(3, 1, i + 1)
            plt.plot(t, RPY_leaf[i], label='Droneleaf')
            plt.plot(t, RPY_px4[i], label='PX4')
            plt.xlabel("Time (seconds)")
            plt.ylabel(f"{label} (rad)")
            plt.title(f"{label} angle")
            plt.legend()
        if filename:
            plt.savefig(filename)
        else:
            plt.show()
        plt.close()
    
    def plot_signals(self, filename=None):
        if self.common_times is None or self.best_offset is None:
            print("No signal data. Run compute_best_offset first.")
            return
        
        # Check if signals are multidimensional
        is_multidim = False
        num_dims = 1
        
        if hasattr(self.ros_signal_resampled, 'ndim') and self.ros_signal_resampled.ndim > 1:
            is_multidim = True
            num_dims = self.ros_signal_resampled.shape[1]
        elif hasattr(self.ulog_signal_resampled, 'ndim') and self.ulog_signal_resampled.ndim > 1:
            is_multidim = True
            num_dims = self.ulog_signal_resampled.shape[1]
        
        plt.figure(figsize=(10, 3 * num_dims) if is_multidim else (10, 6))
        
        if is_multidim:
            for i in range(num_dims):
                plt.subplot(num_dims, 1, i + 1)
                
                # Extract the i-th dimension from each signal
                ros_dim = self.ros_signal_resampled[:, i] if self.ros_signal_resampled.ndim > 1 else self.ros_signal_resampled
                ulog_dim = self.ulog_signal_resampled[:, i] if self.ulog_signal_resampled.ndim > 1 else self.ulog_signal_resampled
                
                # Apply offset based on which signal is shorter
                if self.shorter_type == 'ROS':
                    # ROS is shorter, shift it by the offset
                    plt.plot(self.common_times + self.best_offset, ros_dim, label='ROS', linewidth=0.5)
                    plt.plot(self.common_times, ulog_dim, label='Ulog', linewidth=0.5)
                else:
                    # Ulog is shorter, shift it by the offset
                    plt.plot(self.common_times, ros_dim, label='ROS', linewidth=0.5)
                    plt.plot(self.common_times + self.best_offset, ulog_dim, label='Ulog', linewidth=0.5)

                plt.ylabel(f"Signal Dim {i}")
                
                if i == 0:  # Only show legend on the first subplot
                    plt.legend()
                
                if i == num_dims - 1:  # Only show x label on the last subplot
                    plt.xlabel("Time (seconds)")
        else:
            # Original code for 1D signals
            if self.shorter_type == 'ROS':
                # ROS is shorter, shift it by the offset
                plt.plot(self.common_times + self.best_offset, self.ros_signal_resampled, label='ROS', linewidth=0.5)
                plt.plot(self.common_times, self.ulog_signal_resampled, label='Ulog', linewidth=0.5)
            else:
                # Ulog is shorter, shift it by the offset
                plt.plot(self.common_times, self.ros_signal_resampled, label='ROS', linewidth=0.5)
                plt.plot(self.common_times + self.best_offset, self.ulog_signal_resampled, label='Ulog', linewidth=0.5)

            plt.xlabel("Time (seconds)")
            plt.ylabel("Signal")
            plt.legend()
        
        plt.suptitle("Resampled Signals")
        plt.tight_layout()
        
        if filename:
            plt.savefig(filename)
        else:
            plt.show()
        plt.close()


class SignalSynchronizerSingle(SignalSynchronizer):
    """
    Synchronizer for single-instance signals.
    This class is used when there is only one instance of the signal in both ROS and ULog.
    """
    def __init__(self, bag_path, ulg_path):
        super().__init__(bag_path, ulg_path)
        self.ros_signal_resampled = None
        self.ulog_signal_resampled = None

    def compute_best_offset(self, ros_signal_class: Union[RosSignal, MotorCommandSignal, CartesianSignal, OrientationSignal],
                            ulog_signal_class: Union[UlogSignal, UlogMotorCommandSignal, UlogCartesianSignal, UlogOrientationSignal],
                            ros_topic_name, ulog_msg_key, 
                            ros_signal_idx=None, ulog_signal_idx=None,
                            **signal_kwargs):
        """
        Compute the time offset between ROS and ULog signals.
        
        Args:
            ros_signal_class: Class to use for ROS signal (e.g., MotorCommandSignal, OrientationSignal)
            ulog_signal_class: Class to use for ULog signal (e.g., UlogMotorCommandSignal, UlogOrientationSignal)
            ros_topic_name: Name of the ROS topic to extract
            ulog_msg_key: Name of the ULog message key to extract
            ros_signal_idx: Index of the signal to use for correlation (None uses all dimensions)
            ulog_signal_idx: Index of the signal to use for correlation (None uses all dimensions)
            **signal_kwargs: Additional keyword arguments to pass to signal classes
        """
        # Load the ROS signal
        with load_rosbag(self.bag_path) as bag:
            ros_signal_obj = ros_signal_class(bag, ros_topic_name, **signal_kwargs)
            
        # Load the ULog signal
        with load_ulg(self.ulg_path) as ulog:
            ulog_signal_obj = ulog_signal_class(ulog, ulog_msg_key, **signal_kwargs)
            
        # Extract timestamps and data
        ros_times = ros_signal_obj.timestamps
        ulog_times = ulog_signal_obj.timestamps
        
        # Extract the specific signal dimension if specified
        if ros_signal_idx is not None:
            ros_signal = ros_signal_obj.get_signal_slice(ros_signal_idx)
        else:
            ros_signal = ros_signal_obj.data
            
        if ulog_signal_idx is not None:
            ulog_signal = ulog_signal_obj.get_signal_slice(ulog_signal_idx)
        else:
            ulog_signal = ulog_signal_obj.data

        self._compute_best_offset(ros_times, ulog_times, ros_signal, ulog_signal)

class SignalSynchronizerMulti(SignalSynchronizer):
    """
    Synchronizer for multi-instance signals.
    This class is used when there are multiple instances of the signal in both ROS and ULog.
    """
    def __init__(self, bag_path, ulg_path):
        super().__init__(bag_path, ulg_path)

    def compute_best_offset(self, ros_signal_class: Union[RosSignal, MotorCommandSignal, CartesianSignal, OrientationSignal],
                            ulog_signal_class: Union[UlogSignal, UlogMotorCommandSignal, UlogCartesianSignal, UlogOrientationSignal],
                            ros_topic_name: str,
                            ulog_msg_keys: List[str],
                            ros_signal_idx=None, ulog_signal_idx=None,
                            **signal_kwargs):
        """
        Compute the time offset between ROS and ULog signals.
        
        Args:
            ros_signal_class: Class to use for ROS signal (e.g., MotorCommandSignal, OrientationSignal)
            ulog_signal_class: Class to use for ULog signal (e.g., UlogMotorCommandSignal, UlogOrientationSignal)
            ros_topic_name: Name of the ROS topic to extract
            ulog_msg_key: Name of the ULog message key to extract
            ros_signal_idx: Index of the signal to use for correlation (None uses all dimensions)
            ulog_signal_idx: Index of the signal to use for correlation (None uses all dimensions)
            **signal_kwargs: Additional keyword arguments to pass to signal classes
        """
        # Load the ROS signal
        with load_rosbag(self.bag_path) as bag:
            ros_signal_obj = ros_signal_class(bag, ros_topic_name, **signal_kwargs)
            
        # Load the ULog signal
        ulog_signal_objs:List[Union[UlogSignal, UlogMotorCommandSignal, UlogCartesianSignal, UlogOrientationSignal]] = []
        with load_ulg(self.ulg_path) as ulog:
            for ulog_msg_key in ulog_msg_keys:
                ulog_signal_objs.append(ulog_signal_class(ulog, ulog_msg_key, **signal_kwargs))

        # Extract timestamps and data
        ros_times = ros_signal_obj.timestamps
        # pick shortest Ulog signal timestamps
        ulog_times = min([ulog_signal_obj.timestamps for ulog_signal_obj in ulog_signal_objs], key=len)

        # turn multiple Ulog signals into a single signal multidimensional array
        n_instances = len(ulog_signal_objs)
        ndim_ulog = ulog_signal_objs[0].dimension
        ndim_ros = ros_signal_obj.dimension

        assert ndim_ulog == ndim_ros, "ROS and ULog signals must have the same number of dimensions"
        
        ulog_signal = np.zeros((len(ulog_times), n_instances * ndim_ulog)) # shape (n_samples, n_instances * ndim_ulog)
        ros_signal = np.zeros((len(ros_times), n_instances * ndim_ros)) # shape (n_samples, n_instances * ndim_ros)
        for instance_idx, ulog_signal_obj in enumerate(ulog_signal_objs):
            if ros_signal_idx is not None:
                ros_signal_i = ros_signal_obj.get_signal_slice(ros_signal_idx)
            else:
                ros_signal_i = ros_signal_obj.data
            
            ros_signal[:, instance_idx * ndim_ros:(instance_idx + 1) * ndim_ros] = ros_signal_i

            if ulog_signal_idx is not None:
                ulog_signal_i = ulog_signal_obj.get_signal_slice(ulog_signal_idx)
            else:
                ulog_signal_i = ulog_signal_obj.data

            ulog_signal[:, instance_idx * ndim_ulog:(instance_idx + 1) * ndim_ulog] = ulog_signal_i[:len(ulog_times)] # Ensure Ulog signal matches ROS time base

        self._compute_best_offset(ros_times, ulog_times, ros_signal, ulog_signal)


from math import sin
import os
import pytest
import numpy as np
from leaf_diag.data_sync import SignalSynchronizer
from leaf_diag.data_loader import OrientationSignal, load_rosbag, load_ulg
from leaf_diag.data_transform import px4_rpy_to_droneleaf_rpy
from leaf_diag.data_loader import (
    OrientationSignal,
    UlogOrientationSignal,
    CartesianSignal,
    UlogPositionSignal,
    BaseSignal
)
from leaf_diag.ulog_export import UlogExporterOutdoor, UlogExporterIndoor
import pandas as pd

import matplotlib.pyplot as plt


@pytest.fixture
def bag_data():
    data_dir = "data_test"
    filename = "2025-03-06-16-39-55.bag"
    path = os.path.join(data_dir, filename)
    topic_name = "/dynamo_db/actu_sys/arm_commander/io_cmds_to_px4"
    return {
        "path": path,
        "topic_name": topic_name
    }

@pytest.fixture
def ulg_data():
    data_dir = "data_test"
    filename = "log_264_2025-3-5-12-17-16.ulg"
    path = os.path.join(data_dir, filename)
    ulog_name = "t_actuator_motors_0"
    motor_id = 3
    expected_offset = 5.796
    return {
        "path": path,
        "ulog_name": ulog_name,
        "motor_id": motor_id,
        "expected_offset": expected_offset
    }

def test_base_signal():
    # Test BaseSignal functionality
    data = np.array([[1, 2], [3, 4], [5, 6]])
    timestamps = np.array([0.0, 0.5, 1.0])
    signal = BaseSignal(data, timestamps)
    
    # Test sampling frequency calculation
    assert np.isclose(signal.sampling_freq, 2.0)
    
    # Test time reset
    signal.reset_time()
    assert np.isclose(signal.timestamps[0], 0.0)
    
    # Test setting start time
    signal.set_start_time(0.5)
    assert np.isclose(signal.timestamps[0], -0.5)
    
    # Test signal slice
    np.testing.assert_array_equal(signal.get_signal_slice(1), np.array([2, 4, 6]))

def test_ros_orientation_signal(bag_data):
    with load_rosbag(bag_data["path"]) as bag:
        ori_signal = OrientationSignal(bag, "/dynamo_db/estimator/ori")
        assert ori_signal.dimension == 3
        assert ori_signal.data_keys == ["roll", "pitch", "yaw"]
        assert ori_signal.data.shape[1] == 3

def test_ulog_orientation_signal(ulg_data):
    with load_ulg(ulg_data["path"]) as ulog:
        ori_signal = UlogOrientationSignal(ulog, "t_vehicle_attitude_0")
        assert ori_signal.dimension == 3
        assert ori_signal.data_keys == ["roll", "pitch", "yaw"]
        assert ori_signal.data.shape[1] == 3

def test_ulog_position_signal(ulg_data):
    with load_ulg(ulg_data["path"]) as ulog:
        pos_signal = UlogPositionSignal(ulog, "t_vehicle_local_position_0")
        assert pos_signal.dimension == 3
        assert pos_signal.data.shape[1] == 3
        
        # Test that PX4 coordinates were transformed to droneleaf coordinates
        # In the transform, x_px4 becomes y_droneleaf, y_px4 becomes x_droneleaf, and z_px4 becomes -z_droneleaf
        
        # Extract original PX4 coordinates from ulog (this is a simplified test)
        orig_x = ulog["t_vehicle_local_position_0"]["f_x"].to_numpy()
        orig_y = ulog["t_vehicle_local_position_0"]["f_y"].to_numpy()
        orig_z = ulog["t_vehicle_local_position_0"]["f_z"].to_numpy()
        
        # Check transformation: y_droneleaf should be x_px4, etc.
        assert np.allclose(pos_signal.data[:, 0], orig_y[:len(pos_signal.data)])
        assert np.allclose(pos_signal.data[:, 1], orig_x[:len(pos_signal.data)])
        assert np.allclose(pos_signal.data[:, 2], -orig_z[:len(pos_signal.data)])

@pytest.mark.parametrize("flight_type,data,single_instance,write_csv,sync_key", [
    (
        "outdoor", 
        {
            "bag_data": "data_test/2025-03-06-16-39-55.bag", 
            "ulg_data": "data_test/log_264_2025-3-5-12-17-16.ulg", 
            "sensor_config" : {"gps" : True, "optical_flow" : True}
        }, 
        False, 
        False,
        "acceleration"
    ),
    (
        "outdoor", 
        {
            "bag_data": None, 
            "ulg_data": "data_test/log_264_2025-3-5-12-17-16.ulg", 
            "sensor_config" : {"gps" : True, "optical_flow" : True}
        }, 
        False, 
        False,
        "acceleration_raw"
    ),
    (
        "outdoor", 
        {
            "bag_data": None, 
            "ulg_data": "data_test/log_264_2025-3-5-12-17-16.ulg", 
            "sensor_config" : {"gps" : False, "optical_flow" : False}
        }, 
        False, 
        False,
        "acceleration_raw"
    ),
    (
        "indoor", 
        {
            "bag_data": "data_test/full_flight.bag", 
            "ulg_data": "data_test/log_137_2025-7-20-14-05-32.ulg"
        }, 
        True, 
        False,
       "acceleration"
    ),
    (
        "indoor", 
        {
            "bag_data": None, 
            "ulg_data": "data_test/log_137_2025-7-20-14-05-32.ulg"
        }, 
        True, 
        False,
       "acceleration"
    ),
    (
        "indoor", 
        {
            "bag_data": "data_test/start_to_midflight.bag", 
            "ulg_data": "data_test/log_139_2025-7-20-14-08-00.ulg"
        }, 
        True, 
        False,
       "acceleration"
    ),
    (
        "indoor", 
        {
            "bag_data": "data_test/midflight_to_land.bag", 
            "ulg_data": "data_test/log_140_2025-7-20-14-08-40.ulg"
        }, 
        True, 
        False,
       "acceleration"
    ),
    (
        "indoor", 
        {
            "bag_data": "data_test/midflight_only.bag", 
            "ulg_data": "data_test/log_141_2025-7-20-14-09-24.ulg"
        }, 
        True, 
        False,
       "acceleration"
    ),
    (
        "outdoor", 
        {
            "bag_data": "data_test/log_25_full-flight.bag", 
            "ulg_data": "data_test/log_25_2025-7-13-16-03-18.ulg", 
            "sensor_config" : {"gps" : True, "optical_flow" : False}
        }, 
        True, 
        False,
        "acceleration_raw"
    ),
    (
        "outdoor", 
        {
            "bag_data": "data_test/log_26_start-to-mid-flight.bag", 
            "ulg_data": "data_test/log_26_2025-7-13-16-04-34.ulg", 
            "sensor_config" : {"gps" : True, "optical_flow" : False}
        }, 
        True, 
        False,
        "acceleration_raw"
    ),
    (
        "outdoor", 
        {
            "bag_data": "data_test/log_27_mid-flight-to-land.bag", 
            "ulg_data": "data_test/log_27_2025-7-13-16-06-36.ulg", 
            "sensor_config" : {"gps" : True, "optical_flow" : False}
        }, 
        True, 
        False,
        "acceleration_raw"
    ),
    (
        "outdoor", 
        {
            "bag_data": "data_test/log_28_mid-flight-only.bag", 
            "ulg_data": "data_test/log_28_2025-7-13-16-07-54.ulg", 
            "sensor_config" : {"gps" : True, "optical_flow" : False}
        }, 
        True, 
        False,
        "acceleration_raw"
    ),
    (
        "outdoor", 
        {
            "bag_data": "data_test/log_29_mid-flight-to-land.bag", 
            "ulg_data": "data_test/log_29_2025-7-13-16-11-20.ulg", 
            "sensor_config" : {"gps" : True, "optical_flow" : False}
        }, 
        True, 
        False,
        "acceleration_raw"
    ),
    (
        "outdoor", 
        {
            "bag_data": "data_test/log_30_mid-flight-only.bag", 
            "ulg_data": "data_test/log_30_2025-7-13-16-14-22.ulg", 
            "sensor_config" : {"gps" : True, "optical_flow" : False}
        }, 
        True, 
        False,
        "acceleration_raw"
    ),
    (
        "outdoor", 
        {
            "bag_data": "data_test/log_31_full-flight.bag", 
            "ulg_data": "data_test/log_31_2025-7-13-16-23-18.ulg", 
            "sensor_config" : {"gps" : True, "optical_flow" : False}
        }, 
        True, 
        False,
        "acceleration_raw"
    ),
    (
        "outdoor", 
        {
            "bag_data": "data_test/log_32_mid-flight-to-land.bag", 
            "ulg_data": "data_test/log_32_2025-7-13-16-24-14.ulg", 
            "sensor_config" : {"gps" : True, "optical_flow" : False}
        }, 
        True, 
        False,
        "acceleration_raw"
    ),
    (
        "outdoor", 
        {
            "bag_data": "data_test/log_32_start-to-mid-flight.bag", 
            "ulg_data": "data_test/log_32_2025-7-13-16-24-14.ulg", 
            "sensor_config" : {"gps" : True, "optical_flow" : False}
        }, 
        True, 
        False,
        "acceleration_raw"
    ),
    (
        "outdoor", 
        {
            "bag_data": "data_test/log_33_mid-flight-to-land.bag", 
            "ulg_data": "data_test/log_33_2025-7-13-16-26-52.ulg", 
            "sensor_config" : {"gps" : True, "optical_flow" : False}
        }, 
        True, 
        False,
        "acceleration_raw"
    ),
], ids=[
    "outdoor_multi", 
    "outdoor_multi_no_rosbag", 
    "outdoor_multi_no_rosbag_no_sensor", 
    "indoor_flight_01", 
    "indoor_flight_01_no_rosbag",
    "indoor_flight_02", 
    "indoor_flight_03",
    "indoor_flight_04", 
    "outdoor_flight_03",
    "outdoor_flight_04",
    "outdoor_flight_05",
    "outdoor_flight_06",
    "outdoor_flight_07",
    "outdoor_flight_08",
    "outdoor_flight_09",
    "outdoor_flight_10a",
    "outdoor_flight_10b",
    "outdoor_flight_11",
])
def test_ulog_exporter(flight_type, data, single_instance, write_csv, sync_key, request):
    bag_data = data["bag_data"]
    ulg_data = data["ulg_data"]
    # get test id
    test_id = request.node.name
    # extract the test id between brackets 'test_ulog_exporter[outdoor_flight_03]'
    test_output_str = test_id.split("[")[1].split("]")[0]
    print(f"Running test: {test_output_str}")

    # Create a temporary output directory
    output_dir = os.path.join("data_test", f"export_test_{test_output_str}")
    os.makedirs(output_dir, exist_ok=True)
    
    # Initialize the exporter based on flight type
    output_bag_name = "test_combined.bag"
    if flight_type == "outdoor":
        exporter = UlogExporterOutdoor(
            rosbag_path=bag_data,
            ulog_path=ulg_data,
            output_dir=output_dir,
            filename=output_bag_name,
            single_instance=single_instance,
            sensor_config=data.get("sensor_config", None)
        )
    else:  # indoor
        exporter = UlogExporterIndoor(
            rosbag_path=bag_data,
            ulog_path=ulg_data,
            output_dir=output_dir,
            filename=output_bag_name,
            single_instance=single_instance
        )
    
    # Test data extraction
    exporter.extract_all_data(sync_key=sync_key, plot_correlation=True)
    
    # verify overlap in synchronizers > 20%
    for topic_name, synchronizer in exporter.synchronizers.items():
        if synchronizer.best_overlap is not None:
            assert synchronizer.best_overlap > 0.5, f"Overlap for {topic_name} is below threshold: {synchronizer.best_overlap}"

    # Verify signals were loaded
    assert exporter.ulog_signals["position"]["signal"] is not None
    assert exporter.ulog_derived_signals["orientation"]["signal"] is not None
    if bag_data:
        assert exporter.ros_signals["position"]["signal"] is not None
        assert exporter.ros_signals["orientation"]["signal"] is not None
        
        # Test CSV export
    csv_output_dir = os.path.join(output_dir, "csv")
    df_dict = exporter.export_df(csv_output_dir, write_csv=write_csv)
    
    # Check if files were created
    assert "px4_position" in df_dict
    assert "px4_orientation" in df_dict
    if write_csv:
        assert os.path.exists(os.path.join(csv_output_dir, "px4_position.csv"))
        assert os.path.exists(os.path.join(csv_output_dir, "px4_orientation.csv"))

    if bag_data:
        assert "ros_position" in df_dict
        assert "ros_orientation" in df_dict
        if write_csv:
            assert os.path.exists(os.path.join(csv_output_dir, "ros_position.csv"))
            assert os.path.exists(os.path.join(csv_output_dir, "ros_orientation.csv"))
            
    # Test orientation angles equivalence after transformation
    px4_ori = exporter.ulog_derived_signals["orientation"]["signal"]
    ros_ori = exporter.ros_signals["orientation"]["signal"] if bag_data else None
    
    # Plot to visually verify the synchronization
    fig, axes = plt.subplots(3, 1, figsize=(10, 8), sharex=True)
    for i, label in enumerate(["Roll", "Pitch", "Yaw"]):
        axes[i].plot(px4_ori.timestamps, px4_ori.data[:, i], 'r-', label=f"PX4 {label}")
        if ros_ori is not None:
            axes[i].plot(ros_ori.timestamps, ros_ori.data[:, i], 'b-', label=f"ROS {label}")
        axes[i].set_ylabel(f"{label} (rad)")
        axes[i].legend()
    
    axes[2].set_xlabel("Time (s)")
    plt.savefig(os.path.join(output_dir, "orientation_comparison.png"))
    plt.close()
    
    # Test position comparison
    px4_pos = exporter.ulog_signals["position"]["signal"]
    ros_pos = exporter.ros_signals["position"]["signal"] if bag_data else None
    
    fig, axes = plt.subplots(3, 1, figsize=(10, 8), sharex=True)
    for i, label in enumerate(["X", "Y", "Z"]):
        axes[i].plot(px4_pos.timestamps, px4_pos.data[:, i], 'r-', label=f"PX4 {label}")
        if ros_pos is not None:
            axes[i].plot(ros_pos.timestamps, ros_pos.data[:, i], 'b-', label=f"ROS {label}")
        axes[i].set_ylabel(f"{label} (m)")
        axes[i].legend()
    
    axes[2].set_xlabel("Time (s)")
    plt.savefig(os.path.join(output_dir, "position_comparison.png"))
    plt.close()
    
    # Check error between ROS and PX4 signals (already synchronized in UlogExporter)
    px4_ori = exporter.ulog_derived_signals["orientation"]["signal"]
    ros_ori = exporter.ros_signals["orientation"]["signal"] if bag_data else None
    px4_pos = exporter.ulog_signals["position"]["signal"]
    ros_pos = exporter.ros_signals["position"]["signal"] if bag_data else None
    
    # Interpolate to compare at exactly the same timestamps
    # For orientation
    if ros_ori is not None:
        highest_frequency = max(px4_ori.sampling_freq, ros_ori.sampling_freq)
        common_ori_timestamps = np.arange(max(px4_ori.timestamps[0], ros_ori.timestamps[0]),
                                        min(px4_ori.timestamps[-1], ros_ori.timestamps[-1]), 1.0 / highest_frequency)
        px4_ori_interp = np.array([np.interp(common_ori_timestamps, px4_ori.timestamps, px4_ori.data[:, i]) for i in range(3)]).T
        ros_ori_interp = np.array([np.interp(common_ori_timestamps, ros_ori.timestamps, ros_ori.data[:, i]) for i in range(3)]).T
        
        # For position
        highest_frequency = max(px4_pos.sampling_freq, ros_pos.sampling_freq)
        common_pos_timestamps = np.arange(max(px4_pos.timestamps[0], ros_pos.timestamps[0]),
                                        min(px4_pos.timestamps[-1], ros_pos.timestamps[-1]), 1.0 / highest_frequency)
        px4_pos_interp = np.array([np.interp(common_pos_timestamps, px4_pos.timestamps, px4_pos.data[:, i]) for i in range(3)]).T
        ros_pos_interp = np.array([np.interp(common_pos_timestamps, ros_pos.timestamps, ros_pos.data[:, i]) for i in range(3)]).T
        
        # Calculate RMSE for orientation (in radians)
        ori_rmse = np.nanmean(np.sqrt(np.square(px4_ori_interp - ros_ori_interp)), axis=0)
        # Calculate RMSE for position (in meters)
        pos_rmse = np.nanmean(np.sqrt(np.square(px4_pos_interp - ros_pos_interp)), axis=0)
        
        # Define acceptable thresholds
        ORI_THRESHOLD = 0.6  # radians (approximately 34 degrees)
        POS_THRESHOLD = 0.5   # meters
        
        # Assert that errors are within acceptable thresholds
        for i, label in enumerate(["Roll", "Pitch", "Yaw"]):
            assert ori_rmse[i] < ORI_THRESHOLD, f"{label} RMSE ({ori_rmse[i]}) exceeds threshold ({ORI_THRESHOLD})"
        
        for i, label in enumerate(["X", "Y", "Z"]):
            assert pos_rmse[i] < POS_THRESHOLD, f"{label} RMSE ({pos_rmse[i]}) exceeds threshold ({POS_THRESHOLD})"
    
    # Test the combined bag creation
    bag_bytes = exporter.create_combined_ulog(output_dir)
    assert os.path.exists(os.path.join(output_dir, output_bag_name))

    # check that the bag contains the following signals
    with load_rosbag(os.path.join(output_dir, output_bag_name)) as output_bag:
        if bag_data:
            ros_messages = [
                "/ros/acceleration",
                "/ros/position",
                "/ros/velocity",
                "/ros/orientation",
                "/ros/orientation_rate"
            ]
            for message in ros_messages:
                df_data = output_bag.message_by_topic_memory(message)
                assert len(df_data) > 0     

        px4_messages = ["/px4/" + value["path"] for value in exporter.ulog_signals.values() if value["signal"] is not None] + \
            ["/px4/" + value["path"] for value in exporter.ulog_derived_signals.values() if value["signal"] is not None] + \
            ["/px4/" + value["path"] + "/" + str(instance_id) for value in exporter.ulog_derived_signals_multi.values() for instance_id in range(value["num_instances"])]

        for message in px4_messages:
            df_data = output_bag.message_by_topic_memory(message)
            assert len(df_data) > 0

    # Test the plotting facility
    plot_dir = "plots"
    exporter.create_charts(plot_dir)
    



import os
from matplotlib import pyplot as plt
import pandas as pd
import json
import io

from .data_loader import (
    MotorCommandSignal,
    load_rosbag, 
    load_ulg, 
    CartesianSignal,
    OrientationSignal,
    UlogPositionSignal,
    UlogVelocitySignal,
    UlogAccelerationSignal,
    UlogRawAccelerationSignal,
    UlogOrientationSignal,
    UlogSignal,
    RosSignal,
    UlogCartesianSignal,
    UlogGPSVelocitySignal,
    UlogAngularSignal,
    ULogGPSPositionSignal,
    UlogMotorCommandSignal,
    UlogOFVelocity,
    UlogHeading,
    UlogEstimatorOFVelocity,
    UlogGyroHeading
)

from .data_sync import SignalSynchronizer, SignalSynchronizerMulti, SignalSynchronizerSingle

from typing import IO, Dict, Union, List, Optional

import rosbag
import rospy

from geometry_msgs.msg import Vector3, Quaternion
from std_msgs.msg import Float32, Float32MultiArray
from tqdm import tqdm

# In ROS, you can write any valid ROS message type to a bag file. Some common message types include:

# 1. Standard primitive messages (`std_msgs`):
#     - `std_msgs/Bool`, `std_msgs/String`
#     - `std_msgs/Int8/16/32/64`, `std_msgs/UInt8/16/32/64`
#     - `std_msgs/Float32`, `std_msgs/Float64`

# 2. Geometry messages:
#     - `geometry_msgs/Point`, `geometry_msgs/Vector3`
#     - `geometry_msgs/Quaternion`, `geometry_msgs/Pose`
#     - `geometry_msgs/PoseStamped`, `geometry_msgs/Transform`
#     - `geometry_msgs/Twist`, `geometry_msgs/Wrench`

# 3. Sensor messages:
#     - `sensor_msgs/Image`, `sensor_msgs/PointCloud2`
#     - `sensor_msgs/Imu`, `sensor_msgs/NavSatFix`
#     - `sensor_msgs/LaserScan`, `sensor_msgs/JointState`

# 4. Navigation messages:
#     - `nav_msgs/Odometry`, `nav_msgs/Path`
#     - `nav_msgs/OccupancyGrid`, `nav_msgs/MapMetaData`

# 5. Custom messages:
#     - Any message type defined in your ROS packages

# Your current function needs to be more flexible to handle different message types based on the data structure.

def add_df_to_bag(
        bag: rosbag.Bag, 
        df: pd.DataFrame,
        topic: str, 
        msg_type_str: str, 
        field_mapping: dict = None
    ):
    """
    Add data from a CSV file to a ROS bag.
    
    Args:
        bag: ROS bag object to write to
        df: DataFrame containing the data to write
        topic: Topic name to publish the data to
        msg_type_str: String identifier for the message type ("vector3", "quaternion", "float32", etc.)
        field_mapping: Dictionary mapping CSV columns to message fields (e.g. {"f_ax": "x", "f_ay": "y", "f_az": "z"})
                     If None, will attempt to match automatically
    Raises:
        Exception: If there is an error writing to the bag or processing the DataFrame
    """
    
    try:
        # Create appropriate message type
        if msg_type_str.lower() == "vector3":
            
            # For acceleration, position, velocity data
            if field_mapping is None:
                # Try to guess mapping based on common patterns
                if all(col in df.columns for col in ['f_ax', 'f_ay', 'f_az']):
                    field_mapping = {'f_ax': 'x', 'f_ay': 'y', 'f_az': 'z'}
                elif all(col in df.columns for col in ['x', 'y', 'z']):
                    field_mapping = {'x': 'x', 'y': 'y', 'z': 'z'}
            
            for _, row in tqdm(df.iterrows(), desc=f"Processing {topic}", total=len(df)):
                msg = Vector3()
                for csv_field, msg_field in field_mapping.items():
                    setattr(msg, msg_field, float(row[csv_field]))
                
                # Convert timestamp to ROS time
                timestamp = rospy.Time.from_sec(float(row['timestamp']))
                bag.write(topic, msg, timestamp)
                
        elif msg_type_str.lower() == "float32_multi_array":
            # For motor commands
            if field_mapping is None:
                # Try to guess mapping based on common patterns
                if all(col in df.columns for col in ['f_control_0_', 'f_control_1_', 'f_control_2_', 'f_control_3_']):
                    field_mapping = {i: f"f_control_{i}_" for i in range(4)}
            
            for _, row in tqdm(df.iterrows(), desc=f"Processing {topic}", total=len(df)):
                msg = Float32MultiArray()
                msg.data = [float(row[col]) for col in field_mapping.keys()]
                
                # Convert timestamp to ROS time
                timestamp = rospy.Time.from_sec(float(row['timestamp']))
                bag.write(topic, msg, timestamp)

        elif msg_type_str.lower() == "float32":
            # For GPS fix type
            if field_mapping is None:
                # Try to guess mapping based on common patterns
                if all(col in df.columns for col in ['f_fix_type']):
                    field_mapping = {'f_fix_type': 'data'}
            
            for _, row in tqdm(df.iterrows(), desc=f"Processing {topic}", total=len(df)):
                msg = Float32()
                for csv_field, msg_field in field_mapping.items():
                    setattr(msg, msg_field, float(row[csv_field]))
                
                # Convert timestamp to ROS time
                timestamp = rospy.Time.from_sec(float(row['timestamp']))
                bag.write(topic, msg, timestamp)

    except Exception as e:
        rospy.logerr(f"Error writing to bag {bag.filename} on topic {topic}: {e}")
        raise e


class BagBuffer(io.BytesIO):
    """BytesIO that looks enough like an open(…, 'wb') file
    for rosbag.Bag to accept it."""
    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self.mode = "wb+"          # rosbag checks this

    def close(self):
        """
        rosbag calls .close() at the very end.
        We only flush to be safe but leave the buffer alive
        so the rest of the code can still read from it.
        """
        if not self.closed:       # protect against double-close
            self.flush()          # make sure everything is written
            # **do not** call super().close() – that would mark it closed

class UlogExporter:
    """
    Class for exporting data from ROS bags and ULog files into a combined format.
    """
    def __init__(self, 
                 rosbag_path: Optional[Union[str, bytes]] = None,
                 ulog_path: Optional[Union[str, bytes]] = None,
                 output_dir="output", 
                 filename="combined_data.bag",
                 single_instance=False):
        """
        Initialize UlogExporter with paths to ROS bag and ULog files.
        
        Args:
            rosbag_path: Path to the ROS bag file
            ulog_path: Path to the ULog file
        """
        self.ulog_signals : Dict[str, UlogSignal | ULogGPSPositionSignal] = {}
        self.ros_signals : Dict[str, RosSignal] = {}
        self.rosbag_path = rosbag_path
        self.ulog_path = ulog_path
        self.single_instance = single_instance

        # Open bag for writing
        self.output_dir = output_dir
        self.output_bag_path = os.path.join(output_dir, filename)

        self.synchronizers: Dict[str, Union[SignalSynchronizer, SignalSynchronizerMulti, SignalSynchronizerSingle]] = {
            # "position" : SignalSynchronizerSingle(bag_path=rosbag_path, ulg_path=ulog_path),
            # "velocity" : SignalSynchronizerSingle(bag_path=rosbag_path, ulg_path=ulog_path),
            "acceleration" : SignalSynchronizerSingle(bag_path=rosbag_path, ulg_path=ulog_path),
            "acceleration_raw" : SignalSynchronizerMulti(bag_path=rosbag_path, ulg_path=ulog_path),
        }

        self.ulog_signals : Dict[str, Dict[str,UlogCartesianSignal|str|Dict[str,str]]] = {
            "position" : {
                "signal" : None, 
                "msg_type" : "vector3", 
                "field_mapping" : {"f_x": "x", "f_y": "y", "f_z": "z"},
                "path" : "vehicle_local_position/position"
            },
            "velocity" : {
                "signal" : None, 
                "msg_type" : "vector3", 
                "field_mapping" : {"f_vx": "x", "f_vy": "y", "f_vz": "z"},
                "path" : "vehicle_local_position/velocity"
            },
            "acceleration" : {
                "signal" : None, 
                "msg_type" : "vector3", 
                "field_mapping" : {"f_ax": "x", "f_ay": "y", "f_az": "z"},
                "path" : "vehicle_local_position/acceleration"
            }
        }

        self.ulog_derived_signals : Dict[str, Dict[str,UlogCartesianSignal|str|Dict[str,str]]] = {
            "orientation" : {
                "signal" : None, 
                "msg_type" : "vector3", 
                "field_mapping" : {"roll": "x", "pitch": "y", "yaw": "z"}, 
                "path" : "vehicle_local_position/orientation"
            },
            "orientation_rate" : {
                "signal" : None, 
                "msg_type" : "vector3", 
                "field_mapping" : {"f_xyz_0_": "x", "f_xyz_1_": "y", "f_xyz_2_": "z"},
                "path" : "vehicle_local_position/orientation_rate"
            },
            "distance_estimate" : {
                "signal" : None, 
                "msg_type" : "float32", 
                "field_mapping" : {"f_dist_bottom": "data"},
                "path" : "vehicle_local_position/distance_estimate"
            },
            "motor_cmds" : {
                "signal" : None, 
                "msg_type" : "float32_multi_array", 
                "field_mapping" : {"f_control_0_": 0, "f_control_1_": 1, "f_control_2_": 2, "f_control_3_": 3},
                "path" : "motor_cmds"
            },
            "sensor_mag": {
                "signal": None,
                "msg_type": "vector3",
                "field_mapping": {"f_x": "x", "f_y": "y", "f_z": "z"},
                "path": "sensor_mag/raw",
            },
            "heading_mag": {
                "signal": None,
                "msg_type": "float32",
                "field_mapping": {"f_heading": "data"},
                "path": "sensor_mag/heading",
            }
        }

        if not self.single_instance:
            self.ulog_derived_signals.update({
                "estimator_status" : {
                    "signal" : None, 
                    "msg_type" : "float32", 
                    "field_mapping" : {"f_primary_instance": "data"},
                    "path" : "estimator_selector_status/primary_instance"
                },
                "n_estimator_instances" : {
                    "signal" : None, 
                    "msg_type" : "float32", 
                    "field_mapping" : {"f_instances_available": "data"},
                    "path" : "estimator_selector_status/num_available_instances"
                },
            })

        self.ulog_derived_signals_multi : Dict[str, Dict[str,List[UlogRawAccelerationSignal|str|Dict[str,str]]]] = {
            "acceleration_raw" : {
                "signal": [],
                "msg_type": "vector3",
                "field_mapping" : {"f_x": "x", "f_y": "y", "f_z": "z"},
                "path" : "sensor_accel/acceleration_raw",
                "num_instances": 0,
                "base_topic_name": "t_sensor_accel_"
            }
        }

        self.ros_signals : Dict[str, Dict[str,CartesianSignal|str|Dict[str,str]]] = {
            "position" : {
                "signal" : None, 
                "msg_type" : "vector3", 
                "field_mapping" : {"x": "x", "y": "y", "z": "z"},
                "path" : "position"
            },
            "velocity" : {
                "signal" : None, 
                "msg_type" : "vector3", 
                "field_mapping" : {"x": "x", "y": "y", "z": "z"},
                "path" : "velocity"
            },
            "acceleration" : {
                "signal" : None, 
                "msg_type" : "vector3", 
                "field_mapping" : {"x": "x", "y": "y", "z": "z"},
                "path" : "acceleration"
            },
            "acceleration_raw" : {
                "signal" : None, 
                "msg_type" : "vector3", 
                "field_mapping" : {"x": "x", "y": "y", "z": "z"},
                "path" : "acceleration_raw"
            },
            "orientation" : {
                "signal" : None, 
                "msg_type" : "vector3", 
                "field_mapping" : {"roll": "x", "pitch": "y", "yaw": "z"},
                "path" : "orientation"
            },
            "orientation_rate" : {
                "signal" : None, 
                "msg_type" : "vector3", 
                "field_mapping" : {"roll": "x", "pitch": "y", "yaw": "z"},
                "path" : "orientation_rate"
            },
        }

        self.num_estimator_instances = 0

        self.custom_synchronizers = {
            "acceleration_raw" : {
                "ros_signal" : {
                    "key": "acceleration",
                    "type": "single"
                },
                "ulog_signal" : {
                    "key": "acceleration_raw",
                    "type": "multi",
                }
            }
        }
            

    def extract_all_data(self):
        """
        Extract all available data from both sources
        Returns:
            None
        Raises:
            AssertionError: If the number of estimator instances in the ULog file does not match the expected number.
            Exception: If the ULog file does not contain the expected topics for position, velocity, or acceleration data.
            RuntimeError: If the ULog file cannot be loaded or does not contain the expected topics.
            FileNotFoundError: If the ULog file does not exist or cannot be opened.
            TypeError: If the ULog file is not a valid ULog file or does not contain the expected topics.
            ImportError: If the necessary modules for loading ULog files or ROS bags are not available.
            OSError: If there is an error while reading the ULog file or ROS bag.
            Exception: If there is an unexpected error while extracting data from the ULog file or ROS bag.
        """

        # Count instances of multi-instance topics in ULog file
        with load_ulg(self.ulog_path) as ulog:
            ulog_keys = list(ulog.keys())
            
            # Process each multi-instance signal type
            for signal_key, signal_config in self.ulog_derived_signals_multi.items():
                base_topic = signal_config.get("base_topic_name").rstrip("0123456789")
            
                # Find all instances by counting matching topics
                num_instances = 0
                for key in ulog_keys:
                    if key.startswith(base_topic) and key[len(base_topic):].isdigit():
                        instance_number = int(key[len(base_topic):])
                        num_instances = max(num_instances, instance_number + 1)
                
                # Update the instance count in the configuration
                self.ulog_derived_signals_multi[signal_key]["num_instances"] = num_instances
                print(f"Found {num_instances} instances of {base_topic}")

        with load_ulg(self.ulog_path) as ulog:
            self.ulog_signals["position"]["signal"] = UlogPositionSignal(ulog, "t_vehicle_local_position_0")
            self.ulog_signals["velocity"]["signal"] = UlogVelocitySignal(ulog, "t_vehicle_local_position_0")
            self.ulog_signals["acceleration"]["signal"] = UlogAccelerationSignal(ulog, "t_vehicle_local_position_0")

        if self.rosbag_path is not None:
            with load_rosbag(self.rosbag_path) as rosbag:
                self.ros_signals["position"]["signal"] = CartesianSignal(rosbag, "/dynamo_db/estimator/pos")
                self.ros_signals["velocity"]["signal"] = CartesianSignal(rosbag, "/dynamo_db/estimator/vel")
                self.ros_signals["acceleration"]["signal"] = CartesianSignal(rosbag, "/dynamo_db/estimator/acc")
                self.ros_signals["orientation"]["signal"] = OrientationSignal(rosbag, "/dynamo_db/estimator/ori")
                self.ros_signals["orientation_rate"]["signal"] = OrientationSignal(rosbag, "/dynamo_db/estimator/ori_rate")


        with load_ulg(self.ulog_path) as ulog:
            self.ulog_derived_signals["orientation"]["signal"] = UlogOrientationSignal(ulog, "t_vehicle_attitude_0")
            self.ulog_derived_signals["orientation_rate"]["signal"] = UlogAngularSignal(ulog, "t_vehicle_angular_velocity_0")
            self.ulog_derived_signals["motor_cmds"]["signal"] = UlogMotorCommandSignal(ulog, "t_actuator_motors_0")
            self.ulog_derived_signals["distance_estimate"]["signal"] = UlogSignal(ulog, "t_vehicle_local_position_0", ["f_dist_bottom",])
            self.ulog_derived_signals["sensor_mag"]["signal"] = UlogSignal(ulog, "t_sensor_mag_0", ["f_x", "f_y", "f_z"], dimension=3)
            self.ulog_derived_signals["heading_mag"]["signal"] = UlogHeading(ulog, "t_sensor_mag_0", "t_vehicle_attitude_0")

            if not self.single_instance:
                self.ulog_derived_signals["estimator_status"]["signal"] = UlogSignal(ulog, "t_estimator_selector_status_0", ["f_primary_instance",])
                self.ulog_derived_signals["n_estimator_instances"]["signal"] = UlogSignal(ulog, "t_estimator_selector_status_0", ["f_instances_available",])

                # get number of estimator instances from n_estimator_instances
                self.num_estimator_instances = self.ulog_derived_signals["n_estimator_instances"]["signal"].get_signal_slice(0)[0]
            else:
                self.num_estimator_instances = 1

            n_signals = 0
            num_instances = self.ulog_derived_signals_multi["acceleration_raw"].get("num_instances", 0)
            base_topic_name = self.ulog_derived_signals_multi["acceleration_raw"].get("base_topic_name", "t_sensor_accel_")
            for i in range(num_instances):
                self.ulog_derived_signals_multi["acceleration_raw"]["signal"].append(
                    UlogRawAccelerationSignal(ulog, f"{base_topic_name}{i}")
                )
                n_signals += 1
            self.ulog_derived_signals_multi["acceleration_raw"]["num_instances"] = n_signals

    def _compute_correlations_custom(self, plot_correlation):
        """
        Compute correlations between ROS and ULog signals.
        
        Args:
            plot_correlation: Whether to plot the correlation results
            
        Returns:
            None
        Raises:
            AssertionError: If the number of estimator instances in the ULog file does not match the expected number.
            Exception: If the ULog file does not contain the expected topics for position, velocity, or acceleration data.
            RuntimeError: If the ULog file cannot be loaded or does not contain the expected topics.
            FileNotFoundError: If the ULog file does not exist or cannot be opened.
            TypeError: If the ULog file is not a valid ULog file or does not contain the expected topics.
            ImportError: If the necessary modules for loading ULog files or ROS bags are not available.
            OSError: If there is an error while reading the ULog file or ROS bag.
            Exception: If there is an unexpected error while computing correlations.
        """

        # compute best offset
        for key, sync_settings in self.custom_synchronizers.items():

            if key not in self.synchronizers:
                continue

            ros_signal = self.ros_signals[sync_settings["ros_signal"]["key"]]
            if sync_settings["ulog_signal"]["type"] == "single":
                ulog_signal = self.ulog_signals[sync_settings["ulog_signal"]["key"]]
            elif sync_settings["ulog_signal"]["type"] == "multi":
                ulog_signal = self.ulog_derived_signals_multi[sync_settings["ulog_signal"]["key"]]

            if ros_signal["signal"] is None or ulog_signal["signal"] is None:
                continue
            
            if sync_settings["ulog_signal"]["type"] == "single":
                sync_settings = {
                    "ros_signal_class" : ros_signal["signal"].__class__,
                    "ulog_signal_class" : ulog_signal["signal"].__class__,
                    "ros_topic_name" : ros_signal["signal"].topic,
                    "ulog_msg_key" : ulog_signal["signal"].msg_key,
                    "ros_signal_idx" : None,
                    "ulog_signal_idx" : None
                }
            elif sync_settings["ulog_signal"]["type"] == "multi":
                sync_settings = {
                    "ros_signal_class" : ros_signal["signal"].__class__,
                    "ulog_signal_class" : ulog_signal["signal"][0].__class__,
                    "ros_topic_name" : ros_signal["signal"].topic,
                    "ulog_msg_keys" : [ulog_signal["signal"][i].msg_key for i in range(ulog_signal["num_instances"])],
                    "ros_signal_idx" : None,
                    "ulog_signal_idx" : None
                }

            self.synchronizers[key].compute_best_offset(**sync_settings)

            if plot_correlation:
                self.synchronizers[key].plot_correlation(os.path.join(self.output_dir, f"{key}_correlation.png"))
                self.synchronizers[key].plot_signals(os.path.join(self.output_dir, f"{key}_signals.png"))

    def _compute_correlations(self, plot_correlation):
        """
        Compute correlations between ROS and ULog signals.
        
        Args:
            plot_correlation: Whether to plot the correlation results
            
        Returns:
            None
        Raises:
            AssertionError: If the number of estimator instances in the ULog file does not match the expected number.
            Exception: If the ULog file does not contain the expected topics for position, velocity, or acceleration data.
            RuntimeError: If the ULog file cannot be loaded or does not contain the expected topics.
            FileNotFoundError: If the ULog file does not exist or cannot be opened.
            TypeError: If the ULog file is not a valid ULog file or does not contain the expected topics.
            ImportError: If the necessary modules for loading ULog files or ROS bags are not available.
            OSError: If there is an error while reading the ULog file or ROS bag.
            Exception: If there is an unexpected error while computing correlations.
        """

        # compute best offset
        for key, ulog_signal in self.ulog_signals.items():
            
            if key not in self.synchronizers:
                continue

            ros_signal = self.ros_signals[key]

            if ros_signal["signal"] is None or ulog_signal["signal"] is None:
                continue

            sync_settings = {
                "ros_signal_class" : ros_signal["signal"].__class__,
                "ulog_signal_class" : ulog_signal["signal"].__class__,
                "ros_topic_name" : ros_signal["signal"].topic,
                "ulog_msg_key" : ulog_signal["signal"].msg_key,
                "ros_signal_idx" : None,
                "ulog_signal_idx" : None
            }
            
            self.synchronizers[key].compute_best_offset(**sync_settings)
            if plot_correlation:
                self.synchronizers[key].plot_correlation(os.path.join(self.output_dir, f"{key}_correlation.png"))
                self.synchronizers[key].plot_signals(os.path.join(self.output_dir, f"{key}_signals.png"))

    def _syncronize_signals(self, sync_key:str):

        assert sync_key in self.synchronizers, f"Sync key {sync_key} not found in synchronizers"

        if self.synchronizers[sync_key].best_offset is not None:
            # check shorter signal type
            if self.synchronizers[sync_key].shorter_type == "Ulog":

                if -self.synchronizers[sync_key].best_offset > 0:

                    ulog_start_time = 0.0
                    ros_start_time = self.synchronizers[sync_key].best_offset

                elif -self.synchronizers[sync_key].best_offset < 0:

                    ulog_start_time = -self.synchronizers[sync_key].best_offset
                    ros_start_time = 0.0

            elif self.synchronizers[sync_key].shorter_type == "ROS":

                if -self.synchronizers[sync_key].best_offset > 0:

                    ulog_start_time = self.synchronizers[sync_key].best_offset
                    ros_start_time = 0.0

                elif -self.synchronizers[sync_key].best_offset < 0:

                    ulog_start_time = 0.0
                    ros_start_time = -self.synchronizers[sync_key].best_offset

            for key, ulog_signal in list(self.ulog_signals.items()) + list(self.ulog_derived_signals.items()):
                if ulog_signal["signal"] is not None:
                    ulog_signal["signal"].set_start_time(ulog_start_time)
            for key, ros_signal in self.ros_signals.items():
                if ros_signal["signal"] is not None:
                    ros_signal["signal"].set_start_time(ros_start_time)
            for ulog_signal_key, ulog_signal_config in self.ulog_derived_signals_multi.items():
                for i in range(ulog_signal_config["num_instances"]):
                    self.ulog_derived_signals_multi[ulog_signal_key]["signal"][i].set_start_time(ulog_start_time)

    def export_df(self, output_dir, write_csv=True) -> Dict[str, pd.DataFrame]:
        """
        Export all signals to CSV files
        Args:
            output_dir: Directory to save the CSV files
            write_csv: Whether to write the DataFrame to CSV files
        Returns:
            df_dict: Dictionary containing DataFrames for each signal
        Raises:
            OSError: If there is an error creating the output directory or writing the CSV files
            Exception: If there is an error exporting the DataFrame to CSV files
        """
        os.makedirs(output_dir, exist_ok=True)
        
        df_dict = {}

        for signal_name, signal_dict in dict(self.ulog_signals, **self.ulog_derived_signals).items():
            signal_object = signal_dict["signal"]
            if signal_object is not None:
                df = pd.DataFrame({
                    "timestamp": signal_object.timestamps,
                    **{key: signal_object.get_signal_slice(i) for i, key in enumerate(signal_object.data_keys)}
                })
                if write_csv:
                    csv_path = os.path.join(output_dir, f"px4_{signal_name}.csv")
                    df.to_csv(csv_path, index=False)
                    print(f"Exported {signal_name} to {csv_path}")
                df_dict["px4_" + signal_name] = df

        for signal_name, signal_dict in self.ulog_derived_signals_multi.items():
            for i in range(signal_dict["num_instances"]):
                signal_object = signal_dict["signal"][i]
                if signal_object is not None:
                    df = pd.DataFrame({
                        "timestamp": signal_object.timestamps,
                        **{key: signal_object.get_signal_slice(i) for i, key in enumerate(signal_object.data_keys)}
                    })
                    if write_csv:
                        csv_path = os.path.join(output_dir, f"px4_{signal_name}_{i}.csv")
                        df.to_csv(csv_path, index=False)
                        print(f"Exported {signal_name}_{i} to {csv_path}")
                    df_dict[f"px4_{signal_name}_{i}"] = df

        for signal_name, signal_dict in self.ros_signals.items():
            signal_object = signal_dict["signal"]
            if signal_object is not None:
                df = pd.DataFrame({
                    "timestamp": signal_object.timestamps,
                    **{key: signal_object.get_signal_slice(i) for i, key in enumerate(signal_object.data_keys)}
                })
                if write_csv:
                    csv_path = os.path.join(output_dir, f"ros_{signal_name}.csv")
                    df.to_csv(csv_path, index=False)
                    print(f"Exported {signal_name} to {csv_path}")
                df_dict["ros_" + signal_name] = df

        return df_dict
    
    def create_combined_ulog(self, output_dir=None, write_csv=True, write_bag=True) -> bytes:
        """
        Create a combined dataset in ULog format
        Note: This is a basic implementation that exports to CSV files 
        with ULog-compatible naming conventions
        Args:
            output_dir: Directory to save the output files
            write_csv: Whether to write the DataFrame to CSV files
            write_bag: Whether to write the combined data to a ROS bag file
        Returns:
            bag_bytes: Bytes of the combined ROS bag
        Raises:
            OSError: If there is an error creating the output directory or writing the CSV files
            Exception: If there is an error exporting the DataFrame to CSV files or writing the bag
        """
        if output_dir is not None:
            os.makedirs(output_dir, exist_ok=True)
        
        # Create a metadata file with combined information
        metadata = {
            "version": "1.0",
            "description": "Combined ROS and ULog data",
            "timestamp": pd.Timestamp.now().isoformat()
        }
        
        if output_dir is not None:
            with open(os.path.join(output_dir, "metadata.json"), "w") as f:
                json.dump(metadata, f, indent=2)
        
        # create an IO buffer for the output bag
        bag_stream = BagBuffer()

        rospy.loginfo("Creating bag: %s", self.output_bag_path)
        with rosbag.Bag(bag_stream, "w") as bag:
                
            # Export all signals
            dfs = self.export_df(output_dir, write_csv=write_csv)

            # Add all CSV files to the output bag
            for signal_name, signal_dict in dict(self.ulog_signals, **self.ulog_derived_signals).items():
                if signal_dict["signal"] is None:
                    continue
                df = dfs["px4_" + signal_name]
                add_df_to_bag(
                    bag, df, 
                    f"/px4/{signal_dict['path']}", 
                    msg_type_str=signal_dict["msg_type"], 
                    field_mapping=signal_dict["field_mapping"]
                )

            for signal_name, signal_dict in self.ulog_derived_signals_multi.items():
                if signal_dict["signal"] is None:
                    continue
                for i in range(signal_dict["num_instances"]):
                    df = dfs[f"px4_{signal_name}_{i}"]
                    add_df_to_bag(
                        bag, df, 
                        f"/px4/{signal_dict['path']}/{i}", 
                        msg_type_str=signal_dict["msg_type"], 
                        field_mapping=signal_dict["field_mapping"]
                    )

            for signal_name, signal_dict in self.ros_signals.items():
                if signal_dict["signal"] is None:
                    continue
                df = dfs["ros_" + signal_name]
                add_df_to_bag(
                    bag, df, 
                    f"/ros/{signal_dict['path']}", 
                    msg_type_str=signal_dict["msg_type"], 
                    field_mapping=signal_dict["field_mapping"]
                )

        bag_stream.seek(0)
        bag_bytes = bag_stream.getvalue()

        # Write the bag to the output directory
        if write_bag:
            with open(self.output_bag_path, 'wb') as f:
                f.write(bag_bytes)
            print(f"Combined bag written to {self.output_bag_path}")

        return bag_bytes

    def create_charts(self, output_dir: str = "charts"):
        """
        Create charts for all signals
        """
        chart_dir = os.path.join(self.output_dir, output_dir)
        os.makedirs(chart_dir, exist_ok=True)
        
        for signal_name, signal_dict in dict(self.ulog_signals, **self.ulog_derived_signals).items():
            signal_object = signal_dict["signal"]
            if signal_object is not None:
                fig, ax = signal_object.plot()
                fig.suptitle(signal_name)
                fig.savefig(os.path.join(chart_dir, f"{signal_name}.png"))
                fig.clf()
                plt.close(fig)
                print(f"Created chart for {signal_name}")

        for signal_name, signal_dict in self.ulog_derived_signals_multi.items():
            if signal_dict["num_instances"] == 0:
                continue
            n_dims = len(signal_dict["field_mapping"])
            fig, ax = plt.subplots(n_dims, 1, figsize=(10, 2 * n_dims))
            colors = ['blue', 'red', 'green', 'purple', 'orange', 'brown', 'pink', 'gray', 'olive', 'cyan']
            for i in range(signal_dict["num_instances"]):
                color_idx = i % len(colors)  # Cycle through colors if more dimensions than colors
                signal_object = signal_dict["signal"][i]
                for j in range(n_dims):
                    signal_object.plot(ax=ax[j], idx=j, color=colors[color_idx], label=f"{signal_name}_{j}_i.{i}")
                
            fig.savefig(os.path.join(chart_dir, f"{signal_name}.png"))
            fig.clf()
            plt.close(fig)
            print(f"Created chart for {signal_name}")

        for signal_name, signal_dict in self.ros_signals.items():
            signal_object = signal_dict["signal"]
            if signal_object is not None:
                fig, ax = signal_object.plot()
                fig.suptitle(signal_name)
                fig.savefig(os.path.join(chart_dir, f"{signal_name}.png"))
                fig.clf()
                plt.close(fig)
                print(f"Created chart for {signal_name}")

class UlogExporterOutdoor(UlogExporter):
    """
    Class for exporting data from ROS bags and ULog files into a combined format.
    """
    def __init__(self, 
                 rosbag_path=None, 
                 ulog_path=None, 
                 output_dir="output", 
                 filename="combined_data.bag",
                 single_instance=False,
                 sensor_config=None):
        
        super().__init__(rosbag_path, ulog_path, output_dir, filename, single_instance)

        self.sensor_config = sensor_config.copy() if sensor_config is not None else {}

        self.ulog_derived_signals.update({
            "gps" : {
                "signal" : None, 
                "msg_type" : "vector3", 
                "field_mapping" : {"f_lat": "x", "f_lon": "y", "f_alt": "z"},
                "path" : "vehicle_global_position/gps"
            },
            "gps_vel" : {
                "signal" : None, 
                "msg_type" : "vector3", 
                "field_mapping" : {"f_vel_e_m_s": "x", "f_vel_n_m_s": "y", "f_vel_d_m_s": "z"},
                "path" : "sensor_gps/gps_vel"
            },
            "gps_fix" : {
                "signal" : None, 
                "msg_type" : "float32", 
                "field_mapping" : {"f_fix_type": "data"},
                "path" : "sensor_gps/gps_fix"
            },
            "gps_heading": {
                "signal": None,
                "msg_type": "float32",
                "field_mapping": {"f_heading": "data"},
                "path": "sensor_gps/heading"
            },
            "gps_heading_accuracy": {
                "signal": None,
                "msg_type": "float32",
                "field_mapping": {"f_heading_accuracy": "data"},
                "path": "sensor_gps/heading_accuracy"
            },
            "gps_satellites_used": {
                "signal": None,
                "msg_type": "float32",
                "field_mapping": {"f_satellites_used": "data"},
                "path": "sensor_gps/satellites_used"
            },
            "gps_alt": {
                "signal": None,
                "msg_type": "float32",
                "field_mapping": {"f_alt": "data"},
                "path": "sensor_gps/alt"
            },
            "gps_hdop": {
                "signal": None,
                "msg_type": "float32",
                "field_mapping": {"f_hdop": "data"},
                "path": "sensor_gps/hdop"
            },
            "gps_vdop": {
                "signal": None,
                "msg_type": "float32",
                "field_mapping": {"f_vdop": "data"},
                "path": "sensor_gps/vdop"
            },
            "gps_eph": {
                "signal": None,
                "msg_type": "float32",
                "field_mapping": {"f_eph": "data"},
                "path": "sensor_gps/eph"
            },
            "gps_epv": {
                "signal": None,
                "msg_type": "float32",
                "field_mapping": {"f_epv": "data"},
                "path": "sensor_gps/epv"
            },
            "gps_vpos": {
                "signal": None,
                "msg_type": "float32",
                "field_mapping": {"f_gps_vpos": "data"},
                "path": "estimator_status/innovation_test_ratios/vpos"
            },
            "gps_vvel": {
                "signal": None,
                "msg_type": "float32",
                "field_mapping": {"f_gps_vvel": "data"},
                "path": "estimator_status/innovation_test_ratios/vvel"
            },
            "distance_sensor" : {
                "signal" : None, 
                "msg_type" : "float32", 
                "field_mapping" : {"f_current_distance": "data"},
                "path" : "sensor_distance/current_distance"
            },
            "of_flow" : {
                "signal": None, 
                "msg_type": "vector3", 
                "field_mapping": {"f_vx": "x", "f_vy": "y", "f_vz": "z"},
                "path" : "leaf_signals/of_flow"
            }
        })

        self.ulog_derived_signals_multi.update({
            "of_flow_estimator" : {
                "signal": [],
                "msg_type": "float32_multi_array",
                "field_mapping" : {"f_vel_ne_0_": 0, "f_vel_ne_1_": 1},
                "path" : "estimator_signals/of_flow_estimator",
                "num_instances": 0,
                "base_topic_name": "t_estimator_optical_flow_vel_"
            },
            "vehicle_magnetometer": {
                "signal": [],
                "msg_type": "vector3",
                "field_mapping": {
                    "f_magnetometer_ga_0_": "x",
                    "f_magnetometer_ga_1_": "y",
                    "f_magnetometer_ga_2_": "z"
                },
                "path": "vehicle_magnetometer",
                "num_instances": 0,
                "base_topic_name": "t_vehicle_magnetometer_"
            },
            "innovation_test_ratios_hpos" : {
                "signal": [],
                "msg_type": "float32_multi_array",
                "field_mapping" : {"f_gps_hpos_0_": 0, "f_gps_hpos_1_": 1},
                "path" : "estimator_status/innovation_test_ratios/hpos",
                "num_instances": 0,
                "base_topic_name": "t_estimator_innovation_test_ratios_"
            },
            "innovation_test_ratios_hvel" : {
                "signal": [],
                "msg_type": "float32_multi_array",
                "field_mapping" : {"f_gps_hvel_0_": 0, "f_gps_hvel_1_": 1},
                "path" : "estimator_status/innovation_test_ratios/hvel",
                "num_instances": 0,
                "base_topic_name": "t_estimator_innovation_test_ratios_"
            },
            "innovation_test_ratios_flow" : {
                "signal": [],
                "msg_type": "float32_multi_array",
                "field_mapping" : {"f_flow_0_": 0, "f_flow_1_": 1},
                "path" : "estimator_status/innovation_test_ratios/flow",
                "num_instances": 0,
                "base_topic_name": "t_estimator_innovation_test_ratios_"
            },
            "sensor_gyro": {
                "signal": [],
                "msg_type": "vector3",
                "field_mapping": {
                    "f_x": "x",
                    "f_y": "y",
                    "f_z": "z"
                },
                "path": "sensor_gyro/angular_velocity",
                "num_instances": 0,
                "base_topic_name": "t_sensor_gyro_"
            },
            "sensor_gyro_integrated": {
                "signal": [],
                "msg_type": "vector3",
                "field_mapping": {
                    "angle_x": "x",
                    "angle_y": "y",
                    "angle_z": "z"
                },
                "path": "sensor_gyro/angular_position",
                "num_instances": 0,
                "base_topic_name": "t_sensor_gyro_"
            }
        })

    def extract_all_data(self, sync_key:str, plot_correlation:bool = False):
        """
        Extract all available data from both sources
        This method overrides the base class method to include GPS and optical flow data extraction.
        It also initializes derived signals for GPS, optical flow, and innovation test ratios.
        It assumes that the ULog file contains the necessary topics for GPS and optical flow data.
        It also checks the sensor configuration to determine which signals to extract.
        Args:
            plot_correlation: Whether to plot correlation charts for the signals
            sync_key: The key for the synchronizer to use for signal synchronization
        Returns:
            None
        Raises:
            AssertionError: If the number of estimator instances in the ULog file does not match the expected number.
            Exception: If the ULog file does not contain the expected topics for GPS or optical flow data.
            KeyError: If the sensor configuration does not contain the expected keys for GPS or optical flow.
            ValueError: If the ULog file does not contain the expected topics for GPS or optical flow data.
            RuntimeError: If the ULog file cannot be loaded or does not contain the expected topics.
            FileNotFoundError: If the ULog file does not exist or cannot be opened.
            TypeError: If the sensor configuration is not a dictionary or does not contain the expected keys.
            ImportError: If the necessary modules for loading ULog files or ROS bags are not available.
            OSError: If there is an error while reading the ULog file or ROS bag.
            Exception: If there is an unexpected error while extracting data from the ULog file or ROS bag.
        """

        super().extract_all_data() # <- calculates and sets num_instances for each ulog_derived_signals_multi

        with load_ulg(self.ulog_path) as ulog:
            if self.sensor_config.get("gps", False):
                self.ulog_derived_signals["gps"]["signal"] = ULogGPSPositionSignal(ulog, "t_vehicle_global_position_0")
                self.ulog_derived_signals["gps_vel"]["signal"] = UlogGPSVelocitySignal(ulog, "t_sensor_gps_0")
                self.ulog_derived_signals["gps_fix"]["signal"] = UlogSignal(ulog, "t_vehicle_gps_position_0", ["f_fix_type",])
                self.ulog_derived_signals["gps_heading"]["signal"] = UlogSignal(ulog, "t_sensor_gps_0", ["f_heading",])
                self.ulog_derived_signals["gps_heading_accuracy"]["signal"] = UlogSignal(ulog, "t_sensor_gps_0", ["f_heading_accuracy",])
                self.ulog_derived_signals["gps_satellites_used"]["signal"] = UlogSignal(ulog, "t_sensor_gps_0", ["f_satellites_used",])
                self.ulog_derived_signals["gps_alt"]["signal"] = UlogSignal(ulog, "t_sensor_gps_0", ["f_alt",])
                self.ulog_derived_signals["gps_hdop"]["signal"] = UlogSignal(ulog, "t_sensor_gps_0", ["f_hdop",])
                self.ulog_derived_signals["gps_vdop"]["signal"] = UlogSignal(ulog, "t_sensor_gps_0", ["f_vdop",])
                self.ulog_derived_signals["gps_eph"]["signal"] = UlogSignal(ulog, "t_sensor_gps_0", ["f_eph",])
                self.ulog_derived_signals["gps_epv"]["signal"] = UlogSignal(ulog, "t_sensor_gps_0", ["f_epv",])
                self.ulog_derived_signals["gps_vpos"]["signal"] = UlogSignal(ulog, "t_estimator_innovation_test_ratios_0", ["f_gps_vpos"])
                self.ulog_derived_signals["gps_vvel"]["signal"] = UlogSignal(ulog, "t_estimator_innovation_test_ratios_0", ["f_gps_vvel"])

            if self.sensor_config.get("optical_flow", False):
                self.ulog_derived_signals["of_flow"]["signal"] = UlogOFVelocity(
                    ulog, "t_sensor_optical_flow_0", "t_vehicle_local_position_0", "t_vehicle_attitude_0", "t_distance_sensor_0"
                )

            n_signals = 0
            if self.sensor_config.get("optical_flow", False):
                num_instances = self.ulog_derived_signals_multi["of_flow_estimator"].get("num_instances", 0)
                base_topic_name = self.ulog_derived_signals_multi["of_flow_estimator"].get("base_topic_name", "t_estimator_optical_flow_vel_")
                for i in range(num_instances):
                    self.ulog_derived_signals_multi["of_flow_estimator"]["signal"].append(
                        UlogEstimatorOFVelocity(ulog, f"{base_topic_name}{i}")
                    )
                    n_signals += 1
            self.ulog_derived_signals_multi["of_flow_estimator"]["num_instances"] = n_signals

            n_signals = 0
            num_instances = self.ulog_derived_signals_multi["vehicle_magnetometer"].get("num_instances", 0)
            base_topic_name = self.ulog_derived_signals_multi["vehicle_magnetometer"].get("base_topic_name", "t_vehicle_magnetometer_")
            for i in range(num_instances):
                self.ulog_derived_signals_multi["vehicle_magnetometer"]["signal"].append(
                    UlogSignal(
                        ulog,
                        f"{base_topic_name}{i}",
                        data_keys=["f_magnetometer_ga_0_", "f_magnetometer_ga_1_", "f_magnetometer_ga_2_"],
                        dimension=3
                    )
                )
                n_signals += 1
            self.ulog_derived_signals_multi["vehicle_magnetometer"]["num_instances"] = n_signals

            n_signals = 0
            num_instances = self.ulog_derived_signals_multi["innovation_test_ratios_hpos"].get("num_instances", 0)
            base_topic_name = self.ulog_derived_signals_multi["innovation_test_ratios_hpos"].get("base_topic_name", "t_estimator_innovation_test_ratios_")
            for i in range(num_instances):
                self.ulog_derived_signals_multi["innovation_test_ratios_hpos"]["signal"].append(
                    UlogSignal(ulog, f"{base_topic_name}{i}", data_keys=["f_gps_hpos_0_", "f_gps_hpos_1_"], dimension=2)
                )
                n_signals += 1
            self.ulog_derived_signals_multi["innovation_test_ratios_hpos"]["num_instances"] = n_signals

            n_signals = 0
            num_instances = self.ulog_derived_signals_multi["innovation_test_ratios_hvel"].get("num_instances", 0)
            base_topic_name = self.ulog_derived_signals_multi["innovation_test_ratios_hvel"].get("base_topic_name", "t_estimator_innovation_test_ratios_")
            for i in range(num_instances):
                self.ulog_derived_signals_multi["innovation_test_ratios_hvel"]["signal"].append(
                    UlogSignal(ulog, f"{base_topic_name}{i}", data_keys=["f_gps_hvel_0_", "f_gps_hvel_1_"], dimension=2)
                )
                n_signals += 1
            self.ulog_derived_signals_multi["innovation_test_ratios_hvel"]["num_instances"] = n_signals

            n_signals = 0
            num_instances = self.ulog_derived_signals_multi["innovation_test_ratios_flow"].get("num_instances", 0)
            base_topic_name = self.ulog_derived_signals_multi["innovation_test_ratios_flow"].get("base_topic_name", "t_estimator_innovation_test_ratios_")
            for i in range(num_instances):
                self.ulog_derived_signals_multi["innovation_test_ratios_flow"]["signal"].append(
                    UlogSignal(ulog, f"{base_topic_name}{i}", data_keys=["f_flow_0_", "f_flow_1_"], dimension=2)
                )
                n_signals += 1
            self.ulog_derived_signals_multi["innovation_test_ratios_flow"]["num_instances"] = n_signals

            if self.sensor_config.get("distance_sensor", False):
                try:
                    self.ulog_derived_signals["distance_sensor"]["signal"] = UlogSignal(ulog, "t_distance_sensor_0", ["f_current_distance",])
                except KeyError:
                    print("Warning: t_distance_sensor_0 not found in ULog file.")

            n_signals = 0
            num_instances = self.ulog_derived_signals_multi["sensor_gyro"].get("num_instances", 0)
            base_topic_name = self.ulog_derived_signals_multi["sensor_gyro"].get("base_topic_name", "t_sensor_gyro_")
            for i in range(num_instances):
                self.ulog_derived_signals_multi["sensor_gyro"]["signal"].append(
                    UlogSignal(
                        ulog,
                        f"{base_topic_name}{i}",
                        data_keys=["f_x", "f_y", "f_z"],
                        dimension=3
                    )
                )
                n_signals += 1
            self.ulog_derived_signals_multi["sensor_gyro"]["num_instances"] = n_signals

            n_signals = 0
            num_instances = self.ulog_derived_signals_multi["sensor_gyro_integrated"].get("num_instances", 0)
            base_topic_name = self.ulog_derived_signals_multi["sensor_gyro_integrated"].get("base_topic_name", "t_sensor_gyro_")
            for i in range(num_instances):
                self.ulog_derived_signals_multi["sensor_gyro_integrated"]["signal"].append(
                    UlogGyroHeading(ulog, f"{base_topic_name}{i}", "t_vehicle_attitude_0")
                )
                n_signals += 1
            self.ulog_derived_signals_multi["sensor_gyro_integrated"]["num_instances"] = n_signals

        # apply GPS calibration based on initial x,y,z position 
        if self.ulog_derived_signals["gps"]["signal"] is not None:
            self.ulog_derived_signals["gps"]["signal"].apply_calibration_offset(
                calibration_offset=self.ulog_signals["position"]["signal"].get_initial_coords()
            )

        self._compute_correlations(plot_correlation)
        self._compute_correlations_custom(plot_correlation)
        self._syncronize_signals(sync_key=sync_key)

class UlogExporterIndoor(UlogExporter):
    """
    Class for exporting data from ROS bags and ULog files into a combined format.
    """
    def __init__(self, 
                 rosbag_path=None, 
                 ulog_path=None, 
                 output_dir="output", 
                 filename="combined_data.bag",
                 single_instance=False):
        
        super().__init__(rosbag_path, ulog_path, output_dir, filename, single_instance)

    def extract_all_data(self, sync_key:str, plot_correlation:bool=False):
        """
        Extract all available data from both sources
        This method overrides the base class method to include indoor-specific signal extraction.
        It assumes that the ULog file contains the necessary topics for indoor navigation data.
        Args:
            plot_correlation: Whether to plot correlation charts for the signals
            sync_key: Key for the synchronizer to use for signal synchronization
        Returns:
            None
        Raises:
            AssertionError: If the number of estimator instances in the ULog file does not match the expected number.
            Exception: If the ULog file does not contain the expected topics for indoor navigation data.
            KeyError: If the sensor configuration does not contain the expected keys for indoor navigation data.
            ValueError: If the ULog file does not contain the expected topics for indoor navigation data.
            RuntimeError: If the ULog file cannot be loaded or does not contain the expected topics.
            FileNotFoundError: If the ULog file does not exist or cannot be opened.
            TypeError: If the sensor configuration is not a dictionary or does not contain the expected keys.
            ImportError: If the necessary modules for loading ULog files or ROS bags are not available.
            OSError: If there is an error while reading the ULog file or ROS bag.
            Exception: If there is an unexpected error while extracting data from the ULog file or ROS bag.
        """
        super().extract_all_data()

        self._compute_correlations(plot_correlation)
        self._compute_correlations_custom(plot_correlation)
        self._syncronize_signals(sync_key=sync_key)



from bagpy import bagreader
from rosbag.bag import BagMessage
import rosbag
import pandas as pd
from io import StringIO
import pathlib
import os, io


"""
Utility functions for working with ROS bags in memory.
This module provides a subclass of `bagreader` that allows for
message extraction without writing to disk, suitable for cloud environments.
This is particularly useful for environments like AWS Lambda where
disk I/O can be slow or limited.
This module is designed to work with ROS bags and provides methods
to extract messages by topic, convert them to pandas DataFrames,
and return them as CSV strings or bytes for easy upload to cloud storage.
"""


class NamedBytesIO(io.BytesIO):
    def __init__(self, data: bytes, name: str = "memory.bag", mode: str = "rb"):
        super().__init__(data)
        self.name = name
        self.mode = mode  # rosbag expects this when given a file-like


def slotvalues(m: BagMessage, slot: str):
    """
    Helper function to extract values and column names from message slots.
    Parameters
    ----------
    m : BagMessage
        The ROS bag message object.
    slot : str
        The slot name to extract values from.
    Returns
    -------
    tuple
        A tuple containing the values and the corresponding slot names.
    Raises
    ------
    AttributeError
        If the slot does not exist in the message.
    """
    vals = getattr(m, slot)
    try:
        slots = vals.__slots__
        varray = []
        sarray = []
        for s in slots:
            vnew, snew = slotvalues(vals, s)       
            if isinstance(snew, list):
                for i, snn in enumerate(snew):
                    sarray.append(slot + '.' + snn)
                    varray.append(vnew[i])
            elif isinstance(snew, str):
                sarray.append(slot + '.' + snew)
                varray.append(vnew)    
                
        return varray, sarray
    except AttributeError:
        return vals, slot

class MemoryBagReader(bagreader):
    """
    Subclass of bagreader that supports reading from a path or in-memory bytes.
    """
    def __init__(self, bagfile, delimiter: str = ",", verbose: bool = False):
        self.delimiter = delimiter
        self.verbose = verbose

        self._buf = None   # keep strong ref if using in-memory file
        self.bagfile = bagfile

        # -------- Path-like input --------
        if isinstance(bagfile, (str, os.PathLike)):
            path = pathlib.Path(bagfile)
            self.filename = path.name
            self.dir = str(path.parent) or "./"
            open_arg = str(path)
            self.reader = rosbag.Bag(open_arg, mode='r')

        # -------- File-like / bytes input --------
        else:
            # Normalize to a BytesIO
            if isinstance(bagfile, (bytes, bytearray, memoryview)):
                buf = NamedBytesIO(bytes(bagfile))
            else:
                buf = bagfile  # assume file-like object

            # Give it a name if missing (best-effort)
            if not hasattr(buf, "name"):
                try:
                    buf.name = "memory.bag"  # type: ignore[attr-defined]
                except Exception:
                    pass

            self._buf = buf
            self._buf.seek(0)  # IMPORTANT: rewind after any prior use

            self.filename = getattr(self._buf, "name", "memory.bag")
            self.dir = "./"

            # Let rosbag scan without relying on on-disk index
            self.reader = rosbag.Bag(self._buf, mode='r',
                                     allow_unindexed=True, skip_index=True)

        # ---- rest unchanged ----
        info = self.reader.get_type_and_topic_info()
        self.topic_tuple = info.topics.values()
        self.topics = list(info.topics.keys())

        self.message_types = [t.msg_type for t in self.topic_tuple]
        self.n_messages    = [t.message_count for t in self.topic_tuple]
        self.frequency     = [t.frequency for t in self.topic_tuple]

        self.topic_table = pd.DataFrame(
            list(zip(self.topics, self.message_types, self.n_messages, self.frequency)),
            columns=['Topics', 'Types', 'Message Count', 'Frequency']
        )

        self.start_time = self.reader.get_start_time()
        self.end_time   = self.reader.get_end_time()

        # For path inputs, derive a datafolder; for streams, use a safe default
        if isinstance(bagfile, (str, os.PathLike)):
            stem = pathlib.Path(bagfile).with_suffix('')  # drop .bag
            self.datafolder = str(stem)
        else:
            self.datafolder = "./memory_bag"

        if self.verbose:
            print(f"[INFO] MemoryBagReader initialized for {self.filename!r} (dir={self.dir})")

    
    def message_by_topic_memory(self, topic: str) -> pd.DataFrame:
        """
        Extract messages from ROS bag by topic name without writing to disk.
        Returns a pandas DataFrame directly.
        
        This method closely follows the logic from the original bagreader.message_by_topic
        but creates a DataFrame directly instead of writing to CSV.
        
        Parameters
        ----------
        topic : str
            Topic from which to extract messages.
            
        Returns
        -------
        pandas.DataFrame
            DataFrame containing the extracted message data.
        """
        msg_list = []
        tstart = None
        tend = None
        time = []
        
        for topic_name, msg, t in self.reader.read_messages(topics=topic, start_time=tstart, end_time=tend):
            time.append(t)
            msg_list.append(msg)

        msgs = msg_list

        if len(msgs) == 0:
            if self.verbose:
                print(f"No data on the topic: {topic}")
            return pd.DataFrame()

        # Set column names from the slots - using the same logic as original bagreader
        cols = ["Time"]
        m0 = msgs[0]
        slots = m0.__slots__
        for s in slots:
            v, s_names = slotvalues(m0, s)
            if isinstance(v, tuple):
                snew_array = [] 
                p = list(range(0, len(v)))
                snew_array = [s_names + "_" + str(pelem) for pelem in p]
                s_names = snew_array
            
            if isinstance(s_names, list):
                for i, s1 in enumerate(s_names):
                    cols.append(s1)
            else:
                cols.append(s_names)
        
        # Create data rows - using the same logic as original bagreader
        data_rows = []
        for i, m in enumerate(msgs):
            slots = m.__slots__
            vals = []
            vals.append(time[i].secs + time[i].nsecs*1e-9)
            for s in slots:
                v, s_names = slotvalues(m, s)
                if isinstance(v, tuple):
                    snew_array = [] 
                    p = list(range(0, len(v)))
                    snew_array = [s_names + "_" + str(pelem) for pelem in p]
                    s_names = snew_array

                if isinstance(s_names, list):
                    for j, s1 in enumerate(s_names):
                        vals.append(v[j])
                else:
                    vals.append(v)
            data_rows.append(vals)
        
        # Create DataFrame directly
        df = pd.DataFrame(data_rows, columns=cols)
        return df
    
    def message_by_topic_csv_memory(self, topic: str) -> str:
        """
        Extract messages and return as CSV string in memory.
        
        Parameters
        ----------
        topic : str
            Topic from which to extract messages.
            
        Returns
        -------
        str
            CSV string representation of the data.
        """
        df = self.message_by_topic_memory(topic)
        if df.empty:
            return ""
        
        # Convert to CSV string
        csv_buffer = StringIO()
        df.to_csv(csv_buffer, index=False, sep=self.delimiter)
        return csv_buffer.getvalue()
    
    def message_by_topic_bytes(self, topic: str) -> bytes:
        """
        Extract messages and return as bytes (for S3 upload).
        
        Parameters
        ----------
        topic : str
            Topic from which to extract messages.
            
        Returns
        -------
        bytes
            CSV data as bytes.
        """
        csv_string = self.message_by_topic_csv_memory(topic)
        return csv_string.encode('utf-8')
    
    def close(self):
        """Close the bag reader."""
        if hasattr(self, 'reader'):
            self.reader.close()
    
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


from typing import Tuple
from scipy.spatial.transform import Rotation as R
import numpy as np
import math

def px4_rpy_to_droneleaf_rpy(rpy_px4: np.ndarray, offset=90) -> np.ndarray:
    """
    convert RPY angles in PX4 frame to RPY angles in droneleaf frame
    roll -> roll
    pitch -> -pitch
    yaw -> -yaw
    """
    degrees_yaw = np.degrees(-rpy_px4[:,2] + math.radians(offset))
    wrapped_yaw = (degrees_yaw + 180) % 360 - 180
    rpy_dl = np.column_stack((rpy_px4[:,0], -rpy_px4[:,1], np.radians(wrapped_yaw)))
    return rpy_dl

def quat_to_euler(q: np.ndarray) -> np.ndarray:
    """
    Convert a quaternion [x, y, z, w] to Euler angles (roll, pitch, yaw) in radians.
    """
    r = R.from_quat(q)
    return r.as_euler('xyz', degrees=False)

def get_rotations_from_attitude(q0: float, q1: float, q2: float, q3: float) -> Tuple[R, R]:
    """
    Compute Rotation objects for (NED->body) and (body->NED).
    
    Returns:
        (rot_ned_to_body, rot_body_to_ned)
    """

    # 1) Build the rotation object from the *NED->body* quaternion
    #    SciPy's `from_quat` expects [x, y, z, w].
    rot_ned_to_body = R.from_quat([q1, q2, q3, q0])

    # 2) Invert it to get body->NED
    rot_body_to_ned = rot_ned_to_body.inv()

    return rot_ned_to_body, rot_body_to_ned

def compute_altitude_range(
    R_to_earth: np.ndarray,  # shape (3,3,n)
    state_pos: np.ndarray,   # shape (3,n) in NED
    terrain_vpos: np.ndarray = None,  
    ekf2_of_n_pos_x: float = 0.145,
    ekf2_of_n_pos_y: float = 0.0,
    ekf2_of_n_pos_z: float = -0.02,
    ekf2_imu_pos_x: float = 0.0,
    ekf2_imu_pos_y: float = 0.0,
    ekf2_imu_pos_z: float = 0.0,
    ekf2_rng_gnd_clear: float = 0.1
) -> np.ndarray:
    """
    Calculate the same 'range' used by PX4 EKF for an optical flow sensor,
    following the snippet:
        height_above_gnd_est = max(_terrain_vpos - _state.pos(2) - pos_offset_earth(2),
                                   max(_params.rng_gnd_clearance, 0.01f));
        range = height_above_gnd_est / _R_to_earth(2, 2);
    ...
    """
    n = R_to_earth.shape[2]
    range_to_ground = np.zeros(n)

    # Sensor vs. IMU offsets in body coords
    flow_pos_body = np.array([ekf2_of_n_pos_x, ekf2_of_n_pos_y, ekf2_of_n_pos_z])
    imu_pos_body  = np.array([ekf2_imu_pos_x,  ekf2_imu_pos_y,  ekf2_imu_pos_z])
    pos_offset_body = flow_pos_body - imu_pos_body

    if terrain_vpos is None:
        # trivial fallback: assume terrain is at z=0 + clearance
        terrain_vpos = state_pos[2, :] + ekf2_rng_gnd_clear

    for i in range(n):
        R = R_to_earth[:, :, i]      # 3x3 rotation
        z_ned = state_pos[2, i]      # vehicle z in NED

        pos_offset_earth = R.dot(pos_offset_body)

        raw_height = terrain_vpos[i] - z_ned - pos_offset_earth[2]
        height_above_gnd_est = max(raw_height, max(ekf2_rng_gnd_clear, 0.01))

        den = R[2, 2]
        if abs(den) < 1e-6:
            range_to_ground[i] = np.nan
        else:
            range_to_ground[i] = height_above_gnd_est / den

    return range_to_ground